//   Copyright (c)  1999,2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/assert.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/NumTheory.H"
#include "CoCoA/utils.H"

#include <algorithm>
using std::swap;
#include <limits>
using std::numeric_limits;
#include <vector>
using std::vector;

namespace CoCoA
{

  // Anonymous namespace for file local "static" variables.
  namespace
  {
    // mask[n%30] == 0 iff n is divisible by at least one of 2, 3 and 5.
    // n+skip[n%30] is the next number after n not divisible by 2, 3 or 5
    // n-fall[n%30] is the largest number below n not divisible by 2, 3 or 5
//???NOT USED???    const unsigned int mask[30] = {0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,1};
    const unsigned int skip[30] = {1,6,5,4,3,2,1,4,3,2,1,2,1,4,3,2,1,2,1,4,3,2,1,6,5,4,3,2,1,2};
    const unsigned int fall[30] = {1,2,1,2,3,4,5,6,1,2,3,4,1,2,1,2,3,4,1,2,1,2,3,4,1,2,3,4,5,6};


    // Function to compute base^exp mod modulus; assumes 0 <= base < modulus.
    // Unclever for exp >= modulus.  modulus need not be prime, but (modulus-1)^2
    // must fit inside an unsigned long.
    // This is an iterative implementation of binary powering; seems to be
    // usefully faster than the more obvious recursive implementation.
    // If modulus=1, always produces 0.  Produces 1 when asked to compute 0^0.
    unsigned long PowerMod(unsigned long base, unsigned long exp, unsigned long modulus)
    {
      CoCoA_ASSERT(modulus != 0);
      if (modulus > NumericLimits<unsigned long>::MaxSquarableInteger())
        CoCoA_ERROR(ERR::NYI, "PowerMod with modulus greater than sqrt of max unsigned long");

      if (modulus == 1) return 0;
      if (exp == 0 || base == 1) return 1;
      if (base == 0) return 0;
      if (exp == 1) return base;
    
      unsigned long ans = 1;
      while (exp > 1)
      {
        if (exp&1) ans = (ans*base)%modulus;
        exp /= 2;
        base = (base*base)%modulus;
      }
      return (ans*base)%modulus;
    }


    // Miller-Rabin strong pseudo-prime test (for machine integers).
    // Tests whether n is a strong pseudo-prime to base b.
    // Assumes 1 < b < n
    bool StrongPseudoPrime(long b, long n)
    {
      if (n < 2 || b%n == 0) return false;

      // Compute q,r such that  n-1 = q*2^r  with q odd.
      long q = n-1;
      int r = 0;
      while ((q&1) == 0)
      {
        ++r;
        q /= 2;
      }

      // Main loop
      const long n1 = n-1;
      long power = PowerMod(b, q, n);
      if (power == n1 || power == 1) return true;
      for (int i=0; i < r; i++)
      {
        power = (power*power)%n;
        if (power == n1) return true;
        if (power == 1) return false;
      }
      return false;
    }

    // This definition is for 64-bit machines, but remains valid for 32-bitters
    // ASSUMES n is less than sqrt of max unsigned long.
    // I do not recall the reference for this definition.
    bool IsSmallPrime(unsigned long n)
    {
      CoCoA_ASSERT(n <= NumericLimits<unsigned long>::MaxSquarableInteger());
      if (n == 0 || n == 2 || n == 3 || n == 5 || n == 7) return true;
      if (n < 11) return false;
      // Exclude the 4 composite pseudo-primes (up to 2^32).
      if (n ==  746331041UL ||
          n == 2840871041UL ||
          n == 3014101261UL ||
          n == 3215031751UL)
        return false;
      return StrongPseudoPrime(2, n) &&
        StrongPseudoPrime(5, n) &&
        StrongPseudoPrime(7, n);
    }


    // Assumes N >> 0.
    // Returns true if N is prime, and false otherwise.
    // Maybe slow for some large N.
    bool LucasTest(const ZZ& N)
    {
      const factorization<ZZ> facpows = factor(N-1);
      const vector<ZZ>& primes = facpows.myFactors;
      const int NumPrimes = primes.size();
      const int Amax = static_cast<int>(floor(2*log(N)*log(N))); // experimentally checked up to 10^9: a better bound is 2.62*log(N)*log(log(N)) for N > 3
      for (int a=2; a <= Amax; ++a)
      {
        if (a==4 || a==8 || a==9 || a==16) continue; // skip some pure powers
        ZZ pwr;
        for (int i=0; i < NumPrimes; ++i)
        {
          pwr = PowerMod(a, (N-1)/primes[i], N);
          if (pwr == 1) break;
        }
        if (pwr != 1) return true;
      }
      return false;
    }


    // This fn may assume that abs(N) > NumericLimits::MaxSquarableInteger<unsigned long>().
    bool IsBigPrime(const ZZ& N)
    {
      if (!IsProbPrime(abs(N))) return false;
      return LucasTest(abs(N));
//       CoCoA_ERROR(ERR::NYI, "IsBigPrime unable to test reliably for primes whose square exceeds max unsigned long");
//       return IsProbPrime(abs(N)); // CHEAT!!  BUG BUG BUG!!!
    }

  } // end of anonymous namespace


  // Compute (base^exp)%modulus  with exp >= 0  and modulus >= 2
  // The case 0^0 yields 1.
  unsigned long PowerMod(const MachineInteger& base, const MachineInteger& exp, const MachineInteger& modulus)
  {
    if (IsNegative(modulus) || abs(modulus) < 2)
      CoCoA_ERROR(ERR::BadArg, "PowerMod(b,e,m):  modulus must be greater than 1");
    if (IsNegative(exp))
      CoCoA_ERROR(ERR::NegExp, "PowerMod(b,e,m)");
    const unsigned long e = AsUnsignedLong(exp);
    if (e == 0) return 1;
    const unsigned long m = AsUnsignedLong(modulus);
    unsigned long b = abs(base)%m;
    if (b == 0) return 0;
    if (IsNegative(base) && e%2 == 1)
      b = m-b;
    return PowerMod(b,e,m); // Call the fn in the anon namespace above.
  }


  ZZ PowerMod(const ZZ& base, const MachineInteger& exponent, const ZZ& modulus)
  {
    if (IsZero(modulus))
      CoCoA_ERROR(ERR::ZeroModulus, "PowerMod(ZZ, MachineInteger, ZZ)");
    if (IsNegative(exponent))
      CoCoA_ERROR(ERR::NegExp, "PowerMod(ZZ, MachineInteger, ZZ)");
    ZZ ans;
    mpz_powm_ui(mpzref(ans), mpzref(base), AsUnsignedLong(exponent), mpzref(modulus));
    return ans;
  }


  ZZ PowerMod(const MachineInteger& base, const ZZ& exp, const ZZ& modulus)
  {
    if (IsZero(modulus))
      CoCoA_ERROR(ERR::ZeroModulus, "PowerMod(MachineInteger, ZZ, ZZ)");
    if (exponent < 0)
      CoCoA_ERROR(ERR::NegExp, "PowerMod(MachineInteger, ZZ, ZZ)");
    return PowerMod(ZZ(base), exp, modulus);
  }


  ZZ PowerMod(const ZZ& base, const ZZ& exponent, const ZZ& modulus)
  {
    if (IsZero(modulus))
      CoCoA_ERROR(ERR::ZeroModulus, "PowerMod(ZZ, ZZ, ZZ)");
    if (exponent < 0)
      CoCoA_ERROR(ERR::NegExp, "PowerMod(ZZ, ZZ, ZZ)");
    ZZ ans;
    mpz_powm(mpzref(ans), mpzref(base), mpzref(exponent), mpzref(modulus));
    return ans;
  }



  // Non-negative gcd of machine integers.
  // Emphasis is on simplicity rather than utmost speed.
  unsigned long gcd(const MachineInteger& a, const MachineInteger& b)
  {
    unsigned long A = abs(a);
    unsigned long B = abs(b);
    // Dispose of the trivial cases first.
    if (A == 0) return B;
    if (B == 0) return A;
    if (A == 1 || B == 1) return 1;
    if (A == B) return A;

    // General case.
    if (A < B) std::swap(A, B);
    while (B != 0)
    {
      A %= B;
      std::swap(A, B);
    }
    return A;
  }

  const ZZ::rtn gcd(const ZZ& A, const MachineInteger& b)
  { if (IsZero(b)) return abs(A); else return ZZ(gcd(A%abs(b), b)); }

  const ZZ::rtn gcd(const MachineInteger& a, const ZZ& B)
  { if (IsZero(a)) return abs(B); else return ZZ(gcd(a, B%abs(a))); }

  const ZZ::rtn gcd(const ZZ& A, const ZZ& B)
  {
    mpz_t ans;
    mpz_init(ans);
    mpz_gcd(ans, mpzref(A), mpzref(B));
    return ans[0];
  }


  // This fn is OK -- it is not obvious, but the longs cannot overflow: for
  // a proof see Theorem 4.3 & the comment after it in Shoup's book
  // "A Computational Introduction to Number Theory and Algebra".
  unsigned long ExtendedEuclideanAlg(long& CofacA, long& CofacB, unsigned long a, unsigned long b)
  {
    if (a < b) return ExtendedEuclideanAlg(CofacB, CofacA, b, a);

    // Now we are sure that a >= b.
    long m00 = 1;
    long m01 = 0;
    long m10 = 0;
    long m11 = 1;

    while (true)
    {
      if (b==0) { CofacA = m00; CofacB = m01; return a; }
      long q = a/b; // cannot overflow because b > 1
      a -= q*b;
      m00 -= q*m10;
      m01 -= q*m11;

      if (a == 0) { CofacA = m10; CofacB = m11; return b; }
      q = b/a;
      b -= q*a;
      m10 -= q*m00;
      m11 -= q*m01;
    }
  }


  unsigned long ExtGcd(long& CofacA, long& CofacB, const MachineInteger& a, const MachineInteger& b)
  {
    const unsigned long g = ExtendedEuclideanAlg(CofacA, CofacB, abs(a), abs(b));
    if (IsNegative(a)) CofacA = -CofacA;
    if (IsNegative(b)) CofacB = -CofacB;
    return g;
  }

  ZZ ExtGcd(ZZ& CofacA, ZZ& CofacB, const ZZ& A, const ZZ& B)
  {
    ZZ ans;
    mpz_gcdext(mpzref(ans), mpzref(CofacA), mpzref(CofacB), mpzref(A), mpzref(B));
    return ans;
  }


  unsigned long InvMod(const MachineInteger& r, const MachineInteger& m)
  {
    if (IsNegative(m) || abs(m) < 2) CoCoA_ERROR(ERR::BadArg, "InvMod(n1,n2):  modulus must be greater than 1");
    const unsigned long modulus = abs(m);
    unsigned long residue = abs(r)%modulus;
    if (residue==0) return residue;
    if (IsNegative(r)) residue = modulus-residue;
    long inverse, junk;
    if (ExtGcd(inverse,junk,residue,modulus) != 1) return 0;
    if (inverse > 0) return inverse;
    return modulus+inverse;
  }

  unsigned long InvMod(const ZZ& r, const MachineInteger& m)
  {
    if (IsNegative(m) || abs(m) < 2) CoCoA_ERROR(ERR::BadArg, "InvMod(N1,n2):  modulus must be greater than 1");
    return InvMod(r%m, m);
  }

  ZZ InvMod(const MachineInteger& r, const ZZ& m)
  {
    if (m < 2) CoCoA_ERROR(ERR::BadArg, "InvMod(n1,N2):  modulus must be greater than 1");
    const ZZ residue(r);
    ZZ ans;
    const int InverseExists = mpz_invert(mpzref(ans), mpzref(residue), mpzref(m));
    if (!InverseExists) ans = 0;
    return ans;
  }

  ZZ InvMod(const ZZ& r, const ZZ& m)
  {
    if (m < 2) CoCoA_ERROR(ERR::BadArg, "InvMod(N1,N2):  modulus must be greater than 1");
    ZZ ans;
    const int InverseExists = mpz_invert(mpzref(ans), mpzref(r), mpzref(m));
    if (!InverseExists) ans = 0;
    return ans;
  }


  unsigned long lcm(const MachineInteger& a, const MachineInteger& b)
  {
    if (IsZero(a) || IsZero(b)) return 0;
    const unsigned long AoverG = abs(a)/gcd(a,b);
    if (abs(b) > numeric_limits<unsigned long>::max()/AoverG)
      CoCoA_ERROR(ERR::ArgTooBig, "lcm(n1,n2)");
    return AoverG*abs(b);
  }

  const ZZ::rtn lcm(const ZZ& A, const MachineInteger& b)
  {
    return lcm(A, ZZ(b));
  }

  const ZZ::rtn lcm(const MachineInteger& a, const ZZ& B)
  {
    return lcm(ZZ(a), B);
  }

  const ZZ::rtn lcm(const ZZ& a, const ZZ& b)
  {
    mpz_t ans;
    mpz_init(ans);
    mpz_lcm(ans, mpzref(a), mpzref(b));
    return ans[0];
  }



  // A dispatcher: for small values (whose square fits in unsigned long)
  // We call IsSmallPrime if possible, otherwise we call IsBigPrime.
  bool IsPrime(const MachineInteger& mi)
  {
    const unsigned long n = abs(mi);
    if (n <= NumericLimits<unsigned long>::MaxSquarableInteger())
      return IsSmallPrime(n);
    return IsBigPrime(ZZ(n));
  }

  // A dispatcher: for potentially large values (with no limit).
  // We call IsSmallPrime if possible, otherwise we call IsBigPrime.
  bool IsPrime(const ZZ& N)
  {
    long n;
    if (convert(n, N) && abs(n) <= NumericLimits<unsigned long>::MaxSquarableInteger())
      return IsSmallPrime(abs(n));
    return IsBigPrime(abs(N));
  }

  // "Probable prime" test.  Currently uses Miller-Rabin.
  bool IsProbPrime(const MachineInteger& n, unsigned int NumIters /*= 25*/)
  {
    // Just call the slow version -- at least will guarantee coherent behaviour
    return IsProbPrime(ZZ(n), NumIters);
  }

  bool IsProbPrime(const ZZ& N, unsigned int NumIters /*= 25*/)
  {
    // CoCoALib regards 0 as being prime
    if (N == 0) return true;
    return mpz_probab_prime_p(mpzref(abs(N)), NumIters);
  }



  // Main loop appears twice: once for small n, and once for large n.
  unsigned long NextPrime(const MachineInteger& mi)
  {
    if (IsNegative(mi) || IsZero(mi)) CoCoA_ERROR(ERR::BadArg, "NextPrime: arg must be positive");
    unsigned long n = abs(mi);

    // Special cases if n < 5
    if (n < 2) return 2;
    if (n == 2) return 3;
    if (n < 5) return 5;

    const unsigned long SmallLimit = NumericLimits<unsigned long>::MaxSquarableInteger();
    unsigned long n30 = n%30;
    while (n <= SmallLimit)
    {
      const unsigned long delta = skip[n30];
      n += delta;
      if (n30 == 29) n30=1;
      else n30 += delta;
      if (IsSmallPrime(n)) return n;
    }
    // n now exceeds SmallLimit
    while (true)
    {
      const unsigned long delta = skip[n30];
      if (n+delta < n) break; // break if n overflows
      n += delta;
      if (n30 == 29) n30=1;
      else n30 += delta;
      if (IsPrime(n)) return n;
    }
    // Reach here only if "overflow" has occurred.
    return 0; // to signify "overflow"
  }


  // Main loop appears twice: once for small n, and once for large n.
  unsigned long PrevPrime(const MachineInteger& mi)
  {
    if (IsNegative(mi) || IsZero(mi)) CoCoA_ERROR(ERR::BadArg, "PrevPrime: arg must be positive");
    unsigned long n = abs(mi);

    // Special cases if n < 8
    if (n < 3) return 0; // 0 indicates no previous prime
    if (n == 3) return 2;
    if (n < 6) return 3;
    if (n < 8) return 5;

    const unsigned long SmallLimit = NumericLimits<unsigned long>::MaxSquarableInteger();
    unsigned long n30 = n%30;
    while (n > SmallLimit)
    {
      const unsigned long delta = fall[n30];
      n -= delta;
      if (n30 < delta) n30 += 30;
      n30 -= delta;
      if (IsPrime(n)) return n;
    }
    do
    {
      const unsigned long delta = fall[n30];
      n -= delta;
      if (n30 < delta) n30 += 30;
      n30 -= delta;
    } while (!IsSmallPrime(n));
    return n;
  }


  ZZ::rtn NextProbPrime(ZZ N, long NumIters /*= 25*/)
  {
    if (N <= 0) CoCoA_ERROR(ERR::BadArg, "NextProbPrime: arg must be positive");
    if (NumIters <= 0) CoCoA_ERROR(ERR::BadArg, "NextProbPrime: NumIters must be positive");
    // if N is small use NextPrime
    unsigned long n;
    if (convert(n, N))
    {
      unsigned long ans = NextPrime(n);
      if (ans != 0) return ZZ(ans);
    }

    unsigned long N30 = N%30;
    do
    {
      const unsigned long delta = skip[N30];
      N += delta;
      if (N30 == 29) N30=1;
      else N30 += delta;
    } while (!IsProbPrime(N, NumIters));
    return N;
  }

  ZZ::rtn PrevProbPrime(ZZ N, unsigned int NumIters /*= 25*/)
  {
    if (N <= 0) CoCoA_ERROR(ERR::BadArg, "PrevProbPrime: arg must be positive");
    if (NumIters <= 0) CoCoA_ERROR(ERR::BadArg, "PrevProbPrime: NumIters must be positive");
    // if N is small use PrevPrime
    unsigned long n;
    if (convert(n, N))
      return ZZ(PrevPrime(n));

    unsigned long N30 = N%30;
    do
    {
      const unsigned long delta = fall[N30];
      N -= delta;
      if (N30 < delta) N30 += 30;
      N30 -= delta;
    } while (!IsProbPrime(N, NumIters));
    return N;
  }


// PrimeSource PrimeSourceCtor(void)
// {
//   PrimeSource ans;

//   ans = (PrimeSource)MALLOC(sizeof(struct PrimeSourceStruct));
//   ans->start = nextprime(MAX_PRIME/3); /* heuristic/guess */
//   ans->p = ans->start;
//   return ans;
// }

// void PrimeSourceDtor(PrimeSource PS)
// {
//   FREE(PS);
// }

// FFelem NextPrime(PrimeSource PS)
// {
//   if (PS->start > 0)
//   {
//     PS->p = nextprime(PS->p);
//     if (PS->p != 0) return PS->p;
//     PS->p = PS->start;
//     PS->start = 0;
//   }
//   PS->p = prevprime(PS->p);
//   if (PS->p == 0) JERROR(JERROR_PRIMES_EXHAUSTED);
//   return PS->p;
// }


  factorization<long> SmoothFactor(const MachineInteger& n, const MachineInteger& TrialLimit)
  {
    if (!IsSignedLong(TrialLimit) || AsSignedLong(TrialLimit) < 2)
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(n,TrialLimit): pmax must be at least 2 and fit into a machine long");
    // Below Pmax is unsigned long so that the code will work even if input TrialLimit is numeric_limits<long>::max()
    const unsigned long Pmax = AsUnsignedLong(TrialLimit);
    if (!IsSignedLong(n))
      CoCoA_ERROR(ERR::ArgTooBig, "SmoothFactor(n,pmax): number to be factorized must fit into a machine long");
    unsigned long RemainingFactor = abs(n);

    // The code below does the right thing if n==0, namely result is factors=[], exponents=[], RemainingFactor=0

    vector<long> factors;
    vector<long> exponents;
    unsigned long p = 2;
    unsigned long p30 = 5;
    // NB in line below cannot test p*p <= RemainingFactor directly as the product may overflow.
    while (p <= Pmax && p <= RemainingFactor/p)
    {
      int exp = 0;
      unsigned long rem = RemainingFactor % p;
      while (rem == 0)
      {
        ++exp;
        RemainingFactor /= p;
        rem = RemainingFactor % p;
      }
      if (exp > 0)
      {
        factors.push_back(p);
        exponents.push_back(exp);
      }
      // Advance to next p (we avoid multiples of 2,3,5 but do not skip multiples of 7,11,...)
      // Restricting to prime p actually makes the code slower.
      if (p==2) p=3;
      else if (p==3) p=5;
      else
      {
        const unsigned long delta = skip[p30];
        p += delta;
        if (p30>29) p30=1;
        else p30 += delta;
      }
    }
    // if RemainingFactor is below limit, add it to the list of factors found.
    if (RemainingFactor > 1 && RemainingFactor <= Pmax)
    {
      factors.push_back(RemainingFactor);
      exponents.push_back(1);
      RemainingFactor = 1;
    }
    if (IsNegative(n))
      return factorization<long>(factors, exponents, -static_cast<long>(RemainingFactor));
    return factorization<long>(factors, exponents, RemainingFactor);
  }


  factorization<ZZ> SmoothFactor(const ZZ& N, const MachineInteger& TrialLimit)
  {
    if (!IsSignedLong(TrialLimit) || AsSignedLong(TrialLimit) < 2)
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(N,TrialLimit): pmax must be at least 2 and fit into a machine long");
    // Below Pmax is unsigned long so that the code will work even if input TrialLimit is numeric_limits<long>::max()
    const unsigned long Pmax = AsUnsignedLong(TrialLimit);
    ZZ RemainingFactor = abs(N);

    // The code below does the right thing if N==0, namely result is factors=[], exponents=[], RemainingFactor=0

    vector<ZZ> factors;
    vector<long> exponents;
    unsigned long p = 2;
    // NB in line below cannot test p*p <= RemainingFactor directly as the product may overflow.
    while (p <= Pmax && p <= RemainingFactor/p)
    {
      int exp = 0;
      ZZ quo,rem;
      quorem(quo,rem, RemainingFactor, p);
      while (rem == 0)
      {
        ++exp;
        RemainingFactor = quo;
        quorem(quo,rem, RemainingFactor, p);
      }
      if (exp > 0)
      {
        factors.push_back(ZZ(p));
        exponents.push_back(exp);
      }
      p = NextPrime(p);
    }
    // if RemainingFactor is below limit, add it to the list of factors found.
    if (RemainingFactor > 1 && RemainingFactor <= Pmax)
    {
      factors.push_back(RemainingFactor);
      exponents.push_back(1);
      RemainingFactor = 1;
    }
    if (N < 0)
      return factorization<ZZ>(factors, exponents, -RemainingFactor);
    return factorization<ZZ>(factors, exponents, RemainingFactor);
  }

  factorization<ZZ> SmoothFactor(const ZZ& N, const ZZ& TrialLimit)
  {
    if (TrialLimit < 2)
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(N,TrialLimit): TrialLimit must be at least 2");
    
    // Not implemented for large TrialLimit because it would be hideously slow...
    // A naive implementation could simply copy code from SmoothFactor(N,pmax) above.

    long pmax;
    if (!convert(pmax, TrialLimit))
      CoCoA_ERROR(ERR::NYI, "SmoothFactor(N,TrialLimit) with TrialLimit greater than largest signed long");
    return SmoothFactor(N, pmax);
  }


  factorization<long> factor(const MachineInteger& N)
  {
    if (!IsSignedLong(N))
      CoCoA_ERROR(ERR::ArgTooBig, "factor(n): n must fit into a signed long");
    // Simple rather than efficient.
    if (abs(N) < 2) return SmoothFactor(N,2);
    return SmoothFactor(N,N);
  }

  factorization<ZZ> factor(const ZZ& N)
  {
////    GlobalLogput() << "factorizing: " << N << std::endl;
//    const long MaxLong = 2147483647l; // morally numeric_limits<long>::max(), but this may be too big on 64-bitters
    const long PrimeLimit = 10000000;
    factorization<ZZ> ans = SmoothFactor(N, PrimeLimit);
    if (abs(ans.myRemainingFactor) != 1 && (abs(ans.myRemainingFactor) < power(PrimeLimit,2) || IsPrime(abs(ans.myRemainingFactor))))
    { ans.myFactors.push_back(ans.myRemainingFactor); ans.myExponents.push_back(1); ans.myRemainingFactor = 1; return ans;}
    if (abs(ans.myRemainingFactor) != 1)
      CoCoA_ERROR(ERR::NYI, "factor(N) unimplemented in this case -- too many large factors");
    return ans;
  }


  unsigned long EulerPhi(const MachineInteger& n)
  {
    const factorization<long> facpows = factor(n);
    const vector<long>& primes = facpows.myFactors;
    unsigned long ans = abs(n);
    int NumPrimes = primes.size();
    for (int i=0; i < NumPrimes; ++i)
    {
      const unsigned long p = primes[i];
      ans = (ans/p)*(p-1);
    }
    return ans;
  }

  ZZ EulerPhi(const ZZ& N)
  {
    const factorization<ZZ> facpows = factor(N);
    const vector<ZZ>& primes = facpows.myFactors;

    ZZ ans = abs(N);
    int NumPrimes = primes.size();
    for (int i=0; i < NumPrimes; ++i)
    {
      const ZZ p = primes[i];
      ans = (ans/p)*(p-1);
    }
    return ans;
  }




  int PrimitiveRoot(const MachineInteger& pp)
  {
//     CoCoA_ERROR(ERR::NYI, "PrimitiveRoot");
//     return 0;

    unsigned long p = abs(pp);
    // Code below grabbed from SmallPrime.C
    if (p == 2) return 1;
    if (!IsPrime(p))
      CoCoA_ERROR(ERR::BadArg, "PrimitiveRoot requires argument to be prime");
//    if (p > NumericLimits<unsigned long>::MaxSquarableInteger() || !IsPrime(p)) return 0; // failure

//     vector<unsigned long> PrimeFacs; // all integers below 2^32 have <= 9 distinct prime factors
//                                      // all integers below 2^64 have <= 15 distinct prime factors
//     Factorize32bits(PrimeFacs, p-1);
    const factorization<long> facpows = factor(p-1);
//    GlobalLogput() << "facpows=" << facpows << std::endl;
    const vector<long>& primes = facpows.myFactors;
    for (unsigned long root = 2; /*empty*/; ++root)
    {
      CoCoA_ASSERT(root < p);
//      GlobalLogput() << "trying root=" << root << std::endl;
      if (root == 4 || root == 8 || root == 9) continue; // skip the first few prime powers
      bool failed = false;
      for (vector<long>::const_iterator it = primes.begin(); it != primes.end(); ++it)
      {
//        GlobalLogput() << "PowerMod(" << root << ", " << (p-1)/(*it) << ", " << p << ")=" << PowerMod(root, (p-1)/(*it), p) << std::endl;
	if (PowerMod(root, (p-1)/(*it), p) == 1)  // need brackets around *it to avoid "start of comment"
        { failed = true; break; } // effectively a "continue;" for the outer loop
      }
      if (!failed) return root;
    }
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/NumTheory.C,v 1.1 2009/06/05 12:14:55 abbott Exp $
// $Log: NumTheory.C,v $
// Revision 1.1  2009/06/05 12:14:55  abbott
// Major change:
//   created new files NumTheory.H/C  which contain basic number theory operations
//   removed several basic number theory operations from ZZ.H/C
//   removed gcd from MachineInteger.H/C
//   changed names of some basic fns:
//      IsPPrime -> IsProbPrime
//      invmod -> InvMod    (changed signature too)
//      powermod -> PowerMod  (changed signature too)
//   added new fns
//      NextProbPrime & PrevProbPrime
//   consequent changes to other code and tests and examples
//
//
