//   Copyright (c)  2009  Anna M. Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/PPMonoid.H"
#include "CoCoA/ReductionCog.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/SugarDegree.H"
#include "CoCoA/TmpGPoly.H"
#include "CoCoA/assert.H"

#include <algorithm>
using std::max; // for std sugar
#include <iostream>
//using std::ostream;
#include <memory>
using std::auto_ptr;


namespace CoCoA
{
  std::ostream& operator<<(std::ostream& out, const SugarDegree& s)
  { return s->myOutput(out); }


  SugarDegree::SugarDegree(const SugarDegree& rhs)
  {
    if (rhs.myPtr.get()==0)
      myPtr.reset(0);
    else
      myPtr.reset((rhs.myPtr)->myClone());
  }


  SugarDegree& SugarDegree::operator=(const SugarDegree& rhs)
  {
    if (this == &rhs) return *this;
    if (rhs.myPtr.get()==0)
      myPtr.reset(0);
    else
      myPtr.reset((rhs.myPtr)->myClone());
    return *this;
  }
  

  void SugarDegreeBase::myUpdate(ReductionCog F, const GPoly& g)
  { myUpdate(ActiveLPP(F)/LPPForOrd(g), g); }
  

  namespace SugarTypes
  {
    //---  StdDegBase and WDegBase abstract classes -------------

    class StdDegBase: public SugarDegreeBase
    {
    public:
      StdDegBase(long s): SugarDegreeBase() { myValue = s; }
      virtual const degree& myWSugar() const;  ///< returns error
      virtual long myStdSugar() const { return myValue; }
      virtual int myCmp(const SugarDegreeBase& s) const;  // this <=> s ? <0,=0,>0

      virtual std::ostream& myOutput(std::ostream& out) const;
      virtual void myMul(ConstRefPPMonoidElem pp);
      virtual void myUpdate(ConstRefPPMonoidElem CofactorPP, const GPoly& g);
      virtual long myDeg(ConstRefPPMonoidElem pp) const = 0;
    protected:
      long myValue;
    };


    class WDegBase: public SugarDegreeBase
    {
    public:
      WDegBase(const degree& s): SugarDegreeBase(), myValue(s) {}
      virtual const degree& myWSugar() const { return myValue; }
      virtual long myStdSugar() const;   ///< returns error
      virtual int myCmp(const SugarDegreeBase& s) const;  // this <=> s ? <0,=0,>0

      virtual std::ostream& myOutput(std::ostream& out) const;
      virtual void myMul(ConstRefPPMonoidElem pp);
      virtual void myUpdate(ConstRefPPMonoidElem CofactorPP, const GPoly& g);
      virtual degree myDeg(ConstRefPPMonoidElem pp) const = 0;
    protected:
      degree myValue;
    };


    //---  StdDegBase concrete classes -------------

    class StdDegPlain: public StdDegBase
    {
    public:
      StdDegPlain(long s): StdDegBase(s) {}
      StdDegPlain* myClone() const { return new StdDegPlain(myValue); }
      virtual long myDeg(ConstRefPPMonoidElem pp) const;
    };


    class StdDegNoIdx: public StdDegBase
    {
    public:
      StdDegNoIdx(long s, long NoIdx): StdDegBase(s) { myNoIdxValue=NoIdx;}
      StdDegNoIdx* myClone() const { return new StdDegNoIdx(myValue, myNoIdxValue); }
      virtual long myDeg(ConstRefPPMonoidElem pp) const;
    private:
      long myNoIdxValue;
    };


    class StdDegNoIdxSat: public StdDegBase
    {
    public:
      StdDegNoIdxSat(long s, long NoIdx): StdDegBase(s) { myNoIdxValue=NoIdx;}
      StdDegNoIdxSat* myClone() const { return new StdDegNoIdxSat(myValue, myNoIdxValue); }
      virtual long myDeg(ConstRefPPMonoidElem pp) const;
    private:
      long myNoIdxValue;
    };


    class StdDegSat: public StdDegBase
    {
    public:
      StdDegSat(long s): StdDegBase(s) {}
      StdDegSat* myClone() const { return new StdDegSat(myValue); }
      virtual long myDeg(ConstRefPPMonoidElem pp) const;
    };


    //---  WDegBase concrete classes -------------

    class WDegPlain: public WDegBase
    {
    public:
      WDegPlain(const degree& s): WDegBase(s) {}
      WDegPlain* myClone() const { return new WDegPlain(myValue); }
      virtual degree myDeg(ConstRefPPMonoidElem pp) const;
    };


    class WDeg1CompTmp: public WDegBase
    {
    public:
      WDeg1CompTmp(const degree& s): WDegBase(s) {}
      WDeg1CompTmp* myClone() const { return new WDeg1CompTmp(myValue); }
      virtual degree myDeg(ConstRefPPMonoidElem pp) const;
      virtual void myUpdate(ConstRefPPMonoidElem CofactorPP, const GPoly& g); ///< no default
    };


    class WDegConst: public WDegBase
    {
    public:
      WDegConst(const degree& s): WDegBase(s) {}
      WDegConst* myClone() const { return new WDegConst(myValue); }
      virtual degree myDeg(ConstRefPPMonoidElem pp) const;
      virtual void myUpdate(ConstRefPPMonoidElem /*CofactorPP*/, const GPoly& /*g*/) {}; ///< no default
    };


    class WDegSat: public WDegBase ///< wip
    {
    public:
      WDegSat(const degree& s): WDegBase(s) {}
      WDegSat* myClone() const { return new WDegSat(myValue); }
      virtual degree myDeg(ConstRefPPMonoidElem pp) const;
      virtual void myUpdate(ConstRefPPMonoidElem CofactorPP, const GPoly& g);
    };



    //--  StdDegBase  ----------------------------------

    const degree& StdDegBase::myWSugar() const
    {
      CoCoA_ERROR("Wrong type for this sugar", "SugarTypes::StdDegBase");
      static degree NeverUsed(0);
      return NeverUsed;
    }

    std::ostream& StdDegBase::myOutput(std::ostream& out) const
    { return out << myValue; }

    void StdDegBase::myMul(ConstRefPPMonoidElem pp)
    { myValue += myDeg(pp); }

    void StdDegBase::myUpdate(ConstRefPPMonoidElem CofactorPP, const GPoly& g)
    { myValue = max(myValue, myDeg(CofactorPP) + sugar(g)); }


    int StdDegBase::myCmp(const SugarDegreeBase& s) const
    {
      CoCoA_ASSERT((dynamic_cast<const StdDegBase*>(&s)) != 0);
      const StdDegBase& ss = static_cast<const StdDegBase&>(s);
      if (myValue != ss.myValue)
      {
        if (myValue > ss.myValue) return 1; else return -1;
      }
      return 0;
    }


    //--  WDegBase  ----------------------------------

    long WDegBase::myStdSugar() const
    {
      CoCoA_ERROR("Wrong type for this sugar", "SugarTypes::WDegBase");
      return 0;
    }

    std::ostream& WDegBase::myOutput(std::ostream& out) const
    { return out << myValue; }

    void WDegBase::myMul(ConstRefPPMonoidElem pp)
    { myValue += myDeg(pp); }

    void WDegBase::myUpdate(ConstRefPPMonoidElem CofactorPP, const GPoly& g)
    { myValue = top(myValue, myDeg(CofactorPP) + wsugar(g)); }

    int WDegBase::myCmp(const SugarDegreeBase& s) const
    {
      CoCoA_ASSERT(dynamic_cast<const WDegBase*>(&s) != 0);
      const WDegBase& ws = static_cast<const WDegBase&>(s);
      return cmp(myValue, ws.myValue);
    }

    //--  end abstract classes  ----------------------------------

    //--  StdDegPlain  ----------------------------------

    long StdDegPlain::myDeg(ConstRefPPMonoidElem pp) const
    { return StdDeg(pp); }

    //--  StdDegSat  ----------------------------------

    long StdDegSat::myDeg(ConstRefPPMonoidElem pp) const
    { return StdDeg(pp) - exponent(pp, NumIndets(owner(pp))-1); }

    //--  StdDegNoIdx  ----------------------------------

    long StdDegNoIdx::myDeg(ConstRefPPMonoidElem pp) const
    { return StdDeg(pp) - exponent(pp, myNoIdxValue); }

    //--  StdDegNoIdxSat  ----------------------------------

    long StdDegNoIdxSat::myDeg(ConstRefPPMonoidElem pp) const
    {
      return StdDeg(pp)
        - exponent(pp, myNoIdxValue)
        - exponent(pp, NumIndets(owner(pp))-1);
    }

    //--  WDegPlain  ----------------------------------

    degree WDegPlain::myDeg(ConstRefPPMonoidElem pp) const
    { return wdeg(pp); }

    //--  WDegConst  ----------------------------------

    degree WDegConst::myDeg(ConstRefPPMonoidElem pp) const
    { return wdeg(pp); }

    // inline myUpdate: does nothing -- unlike default implementation

    //--  WDeg1CompTmp  ----------------------------------

    degree WDeg1CompTmp::myDeg(ConstRefPPMonoidElem pp) const
    { return wdeg(pp); }

    void WDeg1CompTmp::myUpdate(ConstRefPPMonoidElem CofactorPP, const GPoly& g)
    { // very temporary!!  only for checking backward compatibility
      // moreover it uses max instead of top (a>b ? a : b)
      // could it be useful with only one component?
      if (g.myGRingInfo().IsMyGradingPosPlus())
        CoCoA_ERROR("TMP: use StdDeg for PosTO", "WDeg1CompTmp::myUpdate");
      myValue = max(myValue, wdeg(CofactorPP) + wsugar(g));
    }

    //--  WDegSat  ----------------------------------

    degree WDegSat::myDeg(ConstRefPPMonoidElem pp) const
    {
      // bihomog trick
      degree tmp = wdeg(pp);
      tmp.mySetComponent(0, tmp[1]);//WARNING: this works only for GrDim=1
      return tmp;
    }


    void WDegSat::myUpdate(ConstRefPPMonoidElem CofactorPP, const GPoly& g)
    {
      if (g.myGRingInfo().IsMyGradingPosPlus())
        CoCoA_ERROR("TMP: use StdDeg for PosTO", "WDeg1CompTmp::myUpdate");
      myValue = max(myValue, myDeg(CofactorPP) + wsugar(g));
    }

    //-------

  }  // namespace SugarTypes

  //----------------------------------------------------------------------
  // Here are the pseudo-constructors:

  SugarDegree NewStdSugar(long s)
  { return SugarDegree(new SugarTypes::StdDegPlain(s)); }

  SugarDegree NewStdSugarNoIdx(long s, long NoIdx)
  { return SugarDegree(new SugarTypes::StdDegNoIdx(s, NoIdx)); }

  SugarDegree NewStdSugarSat(long s)
  { return SugarDegree(new SugarTypes::StdDegSat(s)); }

  SugarDegree NewStdSugarNoIdxSat(long s, long NoIdx)
  { return SugarDegree(new SugarTypes::StdDegNoIdxSat(s, NoIdx)); }


  SugarDegree NewWSugar(const degree& s)
  { return SugarDegree(new SugarTypes::WDegPlain(s)); }

  SugarDegree NewWDeg1CompTmp(const degree& s)
  { return SugarDegree(new SugarTypes::WDeg1CompTmp(s)); }

  SugarDegree NewWSugarConst(const degree& s)
  { return SugarDegree(new SugarTypes::WDegConst(s)); }

  SugarDegree NewWSugarSat(const degree& s)
  { return SugarDegree(new SugarTypes::WDegSat(s)); }


} // namespace CoCoA



//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SugarDegree.C,v 1.8 2009/04/24 16:37:41 bigatti Exp $
// $Log: SugarDegree.C,v $
// Revision 1.8  2009/04/24 16:37:41  bigatti
// -- added myDeg and pseudo-ctor
// -- improved inheritance
//
// Revision 1.7  2009/03/20 14:01:34  bigatti
// -- minor changes (names)
//
// Revision 1.6  2009/03/18 16:37:13  bigatti
// -- "almost-final" cleanup, tested against some current sugar definitions
//
// Revision 1.5  2009/03/18 15:12:02  abbott
// Minor cosmetic changes.
//
// Revision 1.4  2009/03/16 16:39:48  bigatti
// -- now use auto_ptr instead of SmartPtrIRC
// -- added copy ctor, assignment, cmp
//
// Revision 1.3  2009/03/16 07:27:20  bigatti
// -- added necessary "const"
// -- added WDeg1CompTmp (temporary, for testing)
//
// Revision 1.2  2009/02/20 15:45:25  bigatti
// -- sugar.H --> SugarDegree.H
//
// Revision 1.1  2009/02/20 13:27:15  bigatti
// -- renamed from "sugar.[CH]"
//
// Revision 1.3  2009/02/20 11:01:15  bigatti
// -- added NewHomogWSugar for graded=homoeneous case (just constant)
//
// Revision 1.2  2009/02/20 09:53:27  bigatti
// -- changed name: sweetener --> SugarDegree
// -- introduced abstract classes SugarTypes::StdDegBase, SugarTypes::WDegBase
//
// Revision 1.1  2009/02/09 13:57:05  bigatti
// -- first import
//
