//   Copyright (c)  2006  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/TmpFactor.H"
#include "CoCoA/RingQ.H"  // for IsQ

// These are from directory TmpFactorDir/
#include "mpz_alias.h"
#include "primes.h"
#include "FF.h"
#include "WGD.h"

// These are from directory TmpFactorDir/multivariate/
#include "DMPZgcd.h"
#include "DMPZfactor.h"
#include "DMPZfactor_modp.h"

/***************************************************************************/
/* CoCoALib includes */

#include "CoCoA/RingHom.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/FractionField.H"

#include <iostream>

/***************************************************************************/

namespace CoCoA
{

  std::ostream& operator<<(std::ostream& out, const FacPow& FP) { return out << "(" << FP.first << ")^" << FP.second; }


// Compute common denom of a poly (but does not clear denoms).
  void CommonDenom(RefRingElem D, ConstRefRingElem f)
  {
    CoCoA_ASSERT(IsSparsePolyRing(owner(f)));
    SparsePolyRing P = AsSparsePolyRing(owner(f));
    CoCoA_ASSERT(IsFractionField(CoeffRing(P)));
    FractionField Q = AsFractionField(CoeffRing(P));
    CoCoA_ASSERT(IsGCDDomain(BaseRing(Q)));
    CoCoA_ASSERT(owner(D) == BaseRing(Q));
    D = 1;
    for (SparsePolyIter it = BeginIter(f); !IsEnded(it); ++it)
    {
      ConstRefRingElem c = coeff(it);
      D = (D*den(c))/gcd(D, den(c));// should be    D = lcm(D, den(c));
    }
  }

// Exception safe???  You're kidding, right?
  DMPZ ConvertToDMPZ(ConstRefRingElem f)
  {
    CoCoA_ASSERT(IsSparsePolyRing(owner(f)));
    SparsePolyRing P = AsSparsePolyRing(owner(f));
    ring R = CoeffRing(P);
    CoCoA_ASSERT(IsFractionField(R));
    CoCoA_ASSERT(IsGCDDomain(BaseRing(AsFractionField(R))));

    const size_t nvars = NumIndets(P);
    DMPZ g = NULL;
    std::vector<long> expv(nvars);
    for (SparsePolyIter it = BeginIter(f); !IsEnded(it); ++it)
    {
      ZZ c;
      if (!IsInteger(c, coeff(it))) CoCoA_ERROR("coeff not in Z","factorize");
      exponents(expv, PP(it));
      int *exps = (int*)malloc(nvars*sizeof(int)); // block absorbed by g
      std::copy(expv.begin(), expv.end(), exps);
      g = DMPZprepend(mpzref(c), exps, g);
    }

    return DMPZreverse(g); /* restore original order of the terms */
  }



/* /\* Converts D*poly to a DMPZ -- D should be the result of arpoly_common_denom *\/ */
/* static DMPZ arpoly2DMPZ_common_denom(arpoly poly, ring R, mpz_t D) */
/* { */
/*   monomial iter; */
/*   int var, *exps; */
/*   mpz_t coeff; */
/*   DMPZ f; */

/*   mpz_init(coeff); */
/*   NVARS = R->indetsNo; */
/*   f = NULL; */
/*   for (iter = poly; iter != NULL; iter = monomial_next(iter)) */
/*   { */
/*     exps = (int*)malloc(NVARS*sizeof(int)); */
/*     for (var = 0; var < NVARS; ++var) */
/*       exps[var] = term_get_true_nth(monomial_term(iter), var+1, R); */
/*     mpz_divexact(coeff, D, mpq_denref((verylongrat)monomial_coeff(iter))); */
/*     mpz_mul(coeff, coeff, mpq_numref((verylongrat)monomial_coeff(iter))); */
/*     f = DMPZprepend(coeff, exps, f); */
/*   } */
/*   mpz_clear(coeff); */
/*   return DMPZreverse(f); /\* restore original order of the terms *\/ */
/* } */


  void ConvertDMPZToRingElem(RefRingElem f, DMPZ poly)
  {
    CoCoA_ASSERT(IsSparsePolyRing(owner(f)));
    SparsePolyRing P = AsSparsePolyRing(owner(f));
    PPMonoid M = PPM(P);
    PPMonoidElem t(M);
    f = 0;
    const size_t nvars = NumIndets(P);
    std::vector<long> exps(nvars);
    for (DMPZ iter = poly; iter; iter = iter->next)
    {
      if (mpz_sgn(iter->coeff) == 0) continue;
      ZZ c(iter->coeff);
      std::copy(iter->exps, iter->exps+nvars, exps.begin());
      f += monomial(P, c, PPMonoidElem(M, exps));
    }
  }

  FactorList ConvertToFactorList(DMPZfactors facs, ring Rx)
  {
    bool HasContent = (mpz_cmp_ui(facs->content, 1) != 0);
    FactorList ans;
    DMPZlist iter;
    int i;
    RingElem factor(Rx);

    for (i=1, iter=facs->list; iter; ++i, iter = iter->next)
    {
      ConvertDMPZToRingElem(factor, iter->poly);
      ans.push_back(std::make_pair(factor, iter->deg));
    }
    if (HasContent) /* BUG BUG NYI */ ;

    DMPZfactors_dtor(facs);  // NASTY - we destroy one of our args!!
    return ans;
  }


// currently discard info about content/commondenom
  FactorList factorize(ConstRefRingElem f)
  {
    CoCoA_ASSERT(true); // assert something
    SparsePolyRing P = AsSparsePolyRing(owner(f));
    FractionField Q = AsFractionField(CoeffRing(P));
    RingHom coeff = CoeffEmbeddingHom(P)(EmbeddingHom(Q));
    RingElem D(BaseRing(AsFractionField(CoeffRing(P))));
    CommonDenom(D, f);
    NVARS = NumIndets(P); // NVARS is a global!!!
    DMPZ g = ConvertToDMPZ(f*coeff(D));
    FactorList ans = ConvertToFactorList(DMPZfactor(g), P); // leaks the factorlist
    DMPZdtor(g);
    return ans;
  }


  // temporary: by Anna 2008-01-25
  bool IsIrredPoly(ConstRefRingElem f)
  {  // ??? warning: factorize syntax might change: now only for multivariate
    FactorList facs;
    PolyRing P(AsPolyRing(owner(f)));
    if (IsConstant(f)) return IsIrred(LC(f));
    // ANNA: should do it for Z as well (taking care of constant factors)
    if (!IsQ(CoeffRing(P)))
      CoCoA_ERROR(ERR::NYI, "IsIrreducible, !IsQ(CoeffRing)");
    if ( IsSparsePolyRing(P) )
      facs = factorize(f);
    else
    {
      PolyRing SPR = NewPolyRing(CoeffRing(AsPolyRing(P)), 1);
      RingHom phi(PolyAlgebraHom(P, SPR, indets(SPR)));
      facs = factorize(phi(f));
    }
    bool NonConstFound = false;
    for (FactorList::const_iterator i = facs.begin(); i != facs.end(); ++i)
      if (!IsConstant(i->first))
      {
        if (NonConstFound || i->second!=1)
          return false;
        NonConstFound = true;
      }
    return true;
  }


} // end of namespace CoCoA

// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/TmpFactor.C,v 1.6 2009/05/14 09:20:03 abbott Exp $
// $Log: TmpFactor.C,v $
// Revision 1.6  2009/05/14 09:20:03  abbott
// Changed a comment.
//
// Revision 1.5  2008/10/07 15:46:10  abbott
// Added missing log/history lines at the ends of a few files.
//
//
