#ifdef CoCoA_WITH_FROBBY
//   Copyright (c)  2008 Bjarke Hammersholt Roune

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for Frobby integration

#include "CoCoA/TmpFrobby.H"

#include "CoCoA/PPMonoid.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/DenseUPolyRing.H"
#include "CoCoA/ideal.H"
#include "CoCoA/ring.H"

#include "gmp.h"

#include "frobby.h"

#include <vector>

#include <iostream>
using namespace std;

namespace CoCoA
{
  namespace
  {
    // Stores the consumed monomials into the passed-in ideal.
    class FrobbyMonomialIdealConsumer : public ::Frobby::IdealConsumer
    {
    public:
      FrobbyMonomialIdealConsumer(const SparsePolyRing& ring): _ring(ring) {}
	  
      virtual void idealBegin(size_t /*varCount*/) { _ideals.push_back(ideal(_ring)); }

      virtual void consume(mpz_ptr* exponentVector)
        {
          const std::size_t count = NumIndets(_ring);

          // TODO: there must be a more efficient way to add a monomial
          // generator to an ideal.
          const std::vector<RingElem>& x = indets(_ring);
          RingElem m(_ring);
          if (count > 0)
            m = power(x[0], 0);
          for (size_t indet = 0; indet < count; ++indet)
            m = m * power(x[indet], ZZ(exponentVector[indet]));

          _ideals.back() += ideal(_ring, m);
        }

      std::vector<ideal>& getIdeals() { return _ideals; }

    private:
      SparsePolyRing _ring;
      std::vector<ideal> _ideals;
    };

    // The purpose of this class is to encapsulate initialization and
    // deallocation of mpz_t, while guaranteeing that the only member
    // variable is an mpz_t, so that an array of WrappedMpzT can be
    // safely cast to an array of mpz_t. mpz_class does not guarantee
    // this. In the absense of this class, it becomes very
    // inconvenient to allocate and initialize an array of mpz_t's in
    // an exception-safe way.
    class WrappedMpzT
    {
    public:
      WrappedMpzT() { mpz_init(_mpz); }

      ~WrappedMpzT() { mpz_clear(_mpz); }

      mpz_t& getMpzT() { return _mpz; }

    private:
      mpz_t _mpz;
    };

    void ToFrobbyIdeal(ideal I, Frobby::Ideal& frobbyIdeal)
    {
      const std::size_t count = NumIndets(AsSparsePolyRing(AmbientRing(I)));
      const std::vector<RingElem>& generators = gens(I);
      for (std::vector<RingElem>::const_iterator it = generators.begin();
           it != generators.end(); ++it)
      {
        for (size_t indet = 0; indet < count; ++indet)
          frobbyIdeal.addExponent(mpzref(ZZExponent(LPP(*it), indet)));
      }
    }
  }

  ideal AlexanderDual(ideal I, ConstRefPPMonoidElem pp)
  {
    SparsePolyRing polyRing = AsSparsePolyRing(AmbientRing(I));
    if (PPM(polyRing) != owner(pp))
      CoCoA_ERROR(ERR::MixedPPMs, "AlexanderDual (Frobby interface)");

    // This is commented out because IsMonomial has been implemented
    // in CoCoA but has not been made public yet. TODO: uncomment this
    // when that becomes possible.
    //
    // if (!IsMonomial(I)) CoCoA_ERROR("Ideal must be monomial.",
    //   "AlexanderDual (Frobby interface)");

    const std::size_t count = NumIndets(polyRing);

    // Convert I to a Frobby ideal.
    Frobby::Ideal frobbyIdeal(count);
    ToFrobbyIdeal(I, frobbyIdeal);

    // Set up the consumer.
    FrobbyMonomialIdealConsumer consumer(polyRing);

    // Set up the point to dualize on. 
    WrappedMpzT* exponentVector = new WrappedMpzT[count];
    try
    {
      for (size_t indep = 0; indep < count; ++indep)
        mpz_set(exponentVector[indep].getMpzT(), mpzref(ZZExponent(pp, indep)));
	  
      // Compute Alexander dual using Frobby.
      Frobby::alexanderDual(frobbyIdeal, (mpz_t*)exponentVector, consumer);
    }
    catch (...)
    {
      delete[] exponentVector;
      throw;
    }

    delete[] exponentVector;

    return consumer.getIdeals().back();
  }

  void IrreducibleDecomposition(ideal I, std::vector<ideal>& components)
  {
    SparsePolyRing polyRing = AsSparsePolyRing(AmbientRing(I));

    // This is commented out because IsMonomial has been implemented
    // in CoCoA but has not been made public yet. TODO: uncomment this
    // when that becomes possible.
    //
    // if (!IsMonomial(I)) CoCoA_ERROR("Ideal must be monomial.",
    //   "IrreducibleDecomposition (Frobby interface)");

    // Convert I to a Frobby::ExternalIdeal.
    Frobby::Ideal frobbyIdeal(NumIndets(polyRing));
    ToFrobbyIdeal(I, frobbyIdeal);

    // Set up the consumer.
    FrobbyMonomialIdealConsumer consumer(polyRing);

    // Compute irreducible decomposition using Frobby
    Frobby::irreducibleDecompositionAsIdeals(frobbyIdeal, consumer);

    // Grab components in an efficient way.
    consumer.getIdeals().swap(components);
  }

  ideal MaximalStandardMonomials(ideal I)
  {
    SparsePolyRing polyRing = AsSparsePolyRing(AmbientRing(I));

    // This is commented out because IsMonomial has been implemented
    // in CoCoA but has not been made public yet. TODO: uncomment this
    // when that becomes possible.
    //
    // if (!IsMonomial(I)) CoCoA_ERROR("Ideal must be monomial.",
    //   "IrreducibleDecomposition (Frobby interface)");

    // Set up the ideal.
    Frobby::Ideal frobbyIdeal(NumIndets(polyRing));
    ToFrobbyIdeal(I, frobbyIdeal);

    // Set up the consumer.
    FrobbyMonomialIdealConsumer consumer(polyRing);

    // Compute the maximal standard monomials using Frobby.
    Frobby::maximalStandardMonomials(frobbyIdeal, consumer);

    return consumer.getIdeals().back();
  }
}

#endif
