//   Copyright (c)  2005,2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/ZZ.H"
#include "CoCoA/convert.H"

//#include<string>
using std::string;
#include <iostream>
using std::ostream;
using std::istream;
using std::ios;

#include<cmath>
using std::log;
using std::pow;
using std::abs;

#include<cstdlib>
using std::abs;
using std::size_t;
#include<limits>
using std::numeric_limits; // used only in mantissa
#include<algorithm>
using std::max;            // used only in mantissa
using std::swap;

#include "CoCoA/error.H"


namespace CoCoA
{

  ZZ::ZZ()
  {
    mpz_init(myRep);
  }


  ZZ::ZZ(MP_INT to_be_owned)
  {
    myRep[0] = to_be_owned;
  }


  ZZ::ZZ(const mpz_t to_be_copied)
  {
    mpz_init_set(myRep, to_be_copied);
  }

// COPY CONSTRUCTOR

  ZZ::ZZ(const ZZ& from)
  {
    mpz_init_set(myRep, from.myRep);
  }


  ZZ::ZZ(const MachineInteger& i)
  {
    if (IsNegative(i))
      mpz_init_set_si(myRep, AsSignedLong(i));
    else
      mpz_init_set_ui(myRep, AsUnsignedLong(i));
  }


  ZZ& ZZ::operator=(const ZZ& rhs)
  {
    if (this == &rhs) return *this;
    mpz_set(myRep, rhs.myRep);
    return *this;
  }



  void swap(ZZ& a, ZZ& b)
  {
    if (&a == &b) return;
    mpz_swap(mpzref(a), mpzref(b));
  }


  ostream& operator<<(ostream& out, const ZZ& N)
  {
    // Dispose of the exceptional case n=0 immediately.
    if (IsZero(N)) return out << '0';
    const int base = (out.flags() & ios::oct) ? 8 :
      (out.flags() & ios::hex) ? 16 : 10;

    // prefix is sign and/or base specification
    string prefix;

    // Firstly put the sign in the prefix...
    if (N < 0) prefix += '-';
    else if (out.flags() & ios::showpos) prefix += '+';

    // Next put base indication in the prefix if required...
    if (out.flags() & ios::showbase)
    {
      if (base == 8 || base == 16) prefix += '0';
      if (base == 16) prefix += 'x';
    }

    // Finally compute the digits...
    string digits;
    if (!convert(digits, N, base))
      return out << "(NUMBER TOO BIG: "
                 << NumDigits(N, base) << " digits base " << base << ")";
    if (N < 0) digits.erase(0,1); // kill second '-' if number is < 0.
    out << prefix << digits;
    return out;
  }



  namespace  // anonymous namespace for file local functions
  {

    // Returns true if c is a valid digit for the given base
    // (both upper and lower cases are allowed for hexadecimal).
    // Note: base must be 8, 10 or 16 (otherwise always gives false)
    bool IsDigitBase(char c, int base) // base = 8, 10 or 16
    {
      switch(base)
      {
          case 10:
            return isdigit(c);
          case 16:
            return isxdigit(c);
          case 8:
            return (isdigit(c) &&  c != '8' && c != '9');
          default:
            return false;
      }
    }
  } // end of anonymous namespace

  istream& operator>>(istream& in, ZZ& N)
  {
    const int base = (in.flags() & ios::oct) ? 8 : (in.flags() & ios::hex) ? 16 : 10;
    if (!in.good()) return in;
    ws(in);

    // Look for sign of number.
    int sign = 0;
    char c = in.peek();
    if (c == '-' || c == '+')
    {
      in.ignore();
      if (c == '-') sign = -1; else sign = 1;
    }

    // Read in as many digits as possible.
    string digits;
    while (in.good())
    {
      c = in.peek();
      if (!IsDigitBase(c, base)) break;
      in.ignore();
      digits += c;
    }
    if (digits.length() > 0)
    {
      if (!convert(N, digits))
        CoCoA_ERROR(ERR::ArgTooBig, "input >> ZZ");
      if (sign < 0) neg(N, N);
      return in;
    }
    if (sign == 0) in.clear(ios::failbit | in.rdstate());
    else in.clear(ios::badbit | in.rdstate()); // lost char indicating sign

    return in;
  }


/////////////////////////////////////////////////////////////////////////////
// STANDARD ARITHMETIC

  const ZZ::rtn operator-(const ZZ& N)
  {
    mpz_t ans;
    mpz_init_set(ans, mpzref(N));
    mpz_neg(ans, ans);
    return ans[0];
  }


  const ZZ::rtn operator+(const ZZ& N1, const ZZ& N2)
  {
    mpz_t ans;
    mpz_init(ans);
    mpz_add(ans, mpzref(N1), mpzref(N2));
    return ans[0];
  }


  const ZZ::rtn operator-(const ZZ& N1, const ZZ& N2)
  {
    mpz_t ans;
    mpz_init(ans);
    mpz_sub(ans, mpzref(N1), mpzref(N2));
    return ans[0];
  }


  const ZZ::rtn operator*(const ZZ& N1, const ZZ& N2)
  {
    mpz_t ans;
    mpz_init(ans);
    mpz_mul(ans, mpzref(N1), mpzref(N2));
    return ans[0];
  }


  const ZZ::rtn operator/(const ZZ& N1, const ZZ& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "ZZ / ZZ");
    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "ZZ / ZZ");
    mpz_t ans;
    mpz_init(ans);
    mpz_fdiv_q(ans, mpzref(N1), mpzref(N2));
    return ans[0];
  }


  const ZZ::rtn operator%(const ZZ& N1, const ZZ& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::ZeroModulus, "ZZ % ZZ");
    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "ZZ %/ ZZ");
    mpz_t ans;
    mpz_init(ans);
    mpz_fdiv_r(ans, mpzref(N1), mpzref(N2));
    return ans[0];
  }


  const ZZ::rtn power(const ZZ& base, const ZZ& exponent)
  {
    if (exponent < 0)
      CoCoA_ERROR(ERR::NegExp, "power(ZZ, ZZ)");
    if (IsZero(base) && IsZero(exponent))
      CoCoA_ERROR(ERR::BadPwrZero, "power(ZZ, ZZ)");
    // Could allow high powers of 0 and +/-1.  Is it worth it?
    unsigned long exp;
    if (!convert(exp, exponent))
      CoCoA_ERROR(ERR::ExpTooBig, "power(ZZ, ZZ)");
    mpz_t ans;
    mpz_init(ans);
    mpz_pow_ui(ans, mpzref(base), exp);
    return ans[0];
  }


  const ZZ::rtn power(const ZZ& base, const MachineInteger& exponent)
  {
    if (IsNegative(exponent))
      CoCoA_ERROR(ERR::NegExp, "power(ZZ, MachineInteger)");
    if (IsZero(base) && IsZero(exponent))
      CoCoA_ERROR(ERR::BadPwrZero, "power(ZZ, MachineInteger)");
    mpz_t ans;
    mpz_init(ans);
    mpz_pow_ui(ans, mpzref(base), AsUnsignedLong(exponent));
    return ans[0];
  }


  const ZZ::rtn power(const MachineInteger& base, const ZZ& exponent)
  {
    if (exponent < 0)
      CoCoA_ERROR(ERR::NegExp, "power(MachineInteger, ZZ)");
    if (IsZero(base) && IsZero(exponent))
      CoCoA_ERROR(ERR::BadPwrZero, "power(MachineInteger, ZZ)");
    // BUG???  Could allow high powers of 0 and 1.  Is it worth it?
    unsigned long exp;
    if (!convert(exp, exponent))
      CoCoA_ERROR(ERR::ExpTooBig, "power(MachineInteger, ZZ)");
    const bool NegateResult = IsNegative(base) & IsOdd(exp);
    const unsigned long B = abs(base);
    mpz_t ans;
    mpz_init(ans);
    mpz_ui_pow_ui(ans, B, exp);
    if (NegateResult) mpz_neg(ans, ans);
    return ans[0];
  }


  const ZZ::rtn power(const MachineInteger& base, const MachineInteger& exponent)
  {
    if (IsNegative(exponent))
      CoCoA_ERROR(ERR::NegExp, "power(MachineInteger, MachineInteger)");
    if (IsZero(base) && IsZero(exponent))
      CoCoA_ERROR(ERR::BadPwrZero, "power(MachineInteger, MachineInteger)");
    const bool NegateResult = IsNegative(base) & IsOdd(AsUnsignedLong(exponent));
    const unsigned long B = abs(base);
    mpz_t ans;
    mpz_init(ans);
    mpz_ui_pow_ui(ans, B, AsUnsignedLong(exponent));
    if (NegateResult) mpz_neg(ans, ans);
    return ans[0];
  }



  const ZZ::rtn random(const ZZ& lo, const ZZ& hi)
  {
    if (lo > hi) CoCoA_ERROR(ERR::BadArg, "random(lo,hi)");
    if (lo == hi) return lo;
    mpz_t rnd, range;
    mpz_init(range);
    mpz_sub(range, mpzref(hi), mpzref(lo));
//    if (mpz_sgn(range) < 0) mpz_neg(range, range);
    mpz_init(rnd);
    mpz_random(rnd, 1+mpz_size(range));
    mpz_add_ui(range, range, 1);
    mpz_mod(rnd, rnd, range);
    mpz_clear(range);
    if (lo > hi) mpz_add(rnd, rnd, mpzref(hi));
    else mpz_add(rnd, rnd, mpzref(lo));
    return rnd[0];
  }

  const ZZ::rtn random(const MachineInteger& lo, const MachineInteger& hi)
  {
    return random(ZZ(lo), ZZ(hi));
  }

  const ZZ::rtn random(const MachineInteger& lo, const ZZ& hi)
  {
    return random(ZZ(lo), hi);
  }

  const ZZ::rtn random(const ZZ& lo, const MachineInteger& hi)
  {
    return random(lo, ZZ(hi));
  }


  // NB halves are rounded towards +infinity.
  const ZZ::rtn round(const ZZ& num, const ZZ& den)
  {
    return (num+den/2) / den; // integer division!!
  }


  bool IsZero(const ZZ& N)
  {
    return mpz_sgn(mpzref(N))==0;
  }


  bool IsOne(const ZZ& N)
  {
    return mpz_cmp_ui(mpzref(N), 1)==0;
  }


  bool IsMinusOne(const ZZ& N)
  {
    return mpz_cmp_si(mpzref(N),-1)==0;
  }


  int sign(const ZZ& N)
  {
    return mpz_sgn(mpzref(N));
  }


  bool IsOdd(const ZZ& N)
  {
    return mpz_odd_p(mpzref(N));
  }


  bool IsEven(const ZZ& N)
  {
    return mpz_even_p(mpzref(N));
  }


  std::size_t NumDigits(const ZZ& N, int base)
  {
    if (base < 2 || base > 36)
      CoCoA_ERROR(ERR::BadNumBase, "NumDigits(ZZ,int)");
    return mpz_sizeinbase(mpzref(N), base);
  }


  double mantissa(const ZZ& N)
  {
    static ZZ mant;
    if (IsZero(N)) return 0.0;
    const long log2N = mpz_sizeinbase(mpzref(N), 2);
    const long w = numeric_limits<long>::digits;
    const long m = numeric_limits<double>::digits;
    const long shift = max(0L, w*((log2N-m)/w)); // pick a shift which is a multiple of wordsize
    mpz_fdiv_q_2exp(mpzref(mant), mpzref(N), shift);
    return mpz_get_d(mpzref(mant))/pow(2.0, int(log2N-shift));
  }


  std::size_t exponent(const ZZ& N)
  {
    return mpz_sizeinbase(mpzref(N), 2);
  }


  double log(const ZZ& N)
  {
    if (IsZero(N))
      CoCoA_ERROR(ERR::BadArg, "log(ZZ)");
    long exp;
    double mantissa = mpz_get_d_2exp(&exp, mpzref(N));
    const double log2 = std::log(2.0);
    return std::log(std::abs(mantissa)) + exp*log2;
//    return std::log(std::abs(mantissa(N))) + exponent(N)*std::log(2.0);//???cannot use plain log, g++ refuses to compile it.
  }


  // This fn comes out a long horrible mess because I cannot return a
  // machine int as a ZZ::rtn.  There must be a design error somewhere.
  ZZ::rtn RangeFactorial(const MachineInteger& lo, const MachineInteger& hi)
  {
    mpz_t ans;
    mpz_init(ans); // set to zero
    // If zero is in the range, result is zero
    if ((IsZero(lo) || IsNegative(lo)) && !IsNegative(hi)) return ans[0];
    mpz_set_ui(ans, 1);
    // First check for empty product.
    if (IsNegative(hi) && !IsNegative(lo)) return ans[0];
    // We now know that hi and lo have the same sign.
    // Two more checks for empty products.
    if (IsNegative(lo) && AsSignedLong(lo) > AsSignedLong(hi)) return ans[0];
    if (!IsNegative(lo) && AsUnsignedLong(lo) > AsUnsignedLong(hi)) return ans[0];
    // we now know that the product is not empty.
    unsigned long l = abs(lo);
    unsigned long h = abs(hi);
    if (IsNegative(hi))
    {
      std::swap(l, h);
      if (((h^l)&1) == 0) mpz_neg(ans, ans);
    }

//    if (hi-lo > 4) return RangeFactorial(lo, (lo+hi)/2)*RangeFactorial(1+(lo+hi)/2, hi);
    // From here on 0 < l < h.
    while (l <= h)
    {
      mpz_mul_ui(ans, ans, l);
      ++l;
    }
    return ans[0];
  }



  // The basic comparsion functions: they return <0, =0, or >0.

  int cmp(const ZZ& N1, const ZZ& N2)
  {
    return mpz_cmp(mpzref(N1), mpzref(N2));
  }


  int cmp(const ZZ& N1, const MachineInteger& n2)
  {
    if (IsNegative(n2))
      return mpz_cmp_si(mpzref(N1), AsSignedLong(n2));
    else
      return mpz_cmp_ui(mpzref(N1), AsUnsignedLong(n2));
  }


  int cmp(const MachineInteger& n1, const ZZ& N2)
  {
    return -cmp(N2, n1);
  }


  // Comparison functions between two ZZs

  bool operator==(const ZZ& N1, const ZZ& N2)
  {
    return cmp(N1, N2) == 0;
  }


  bool operator!=(const ZZ& N1, const ZZ& N2)
  {
    return cmp(N1, N2) != 0;
  }


  bool operator> (const ZZ& N1, const ZZ& N2)
  {
    return cmp(N1, N2) > 0;
  }


  bool operator>=(const ZZ& N1, const ZZ& N2)
  {
    return cmp(N1, N2) >= 0;
  }


  bool operator< (const ZZ& N1, const ZZ& N2)
  {
    return cmp(N1, N2) < 0;
  }


  bool operator<=(const ZZ& N1, const ZZ& N2)
  {
    return cmp(N1, N2) <= 0;
  }


  // Comparison functions between ZZ and MachineInteger

  bool operator==(const ZZ& N1, const MachineInteger& n2)
  {
    return cmp(N1, n2) == 0;
  }


  bool operator!=(const ZZ& N1, const MachineInteger& n2)
  {
    return cmp(N1, n2) != 0;
  }


  bool operator> (const ZZ& N1, const MachineInteger& n2)
  {
    return cmp(N1, n2) > 0;
  }


  bool operator>=(const ZZ& N1, const MachineInteger& n2)
  {
    return cmp(N1, n2) >= 0;
  }


  bool operator< (const ZZ& N1, const MachineInteger& n2)
  {
    return cmp(N1, n2) < 0;
  }


  bool operator<=(const ZZ& N1, const MachineInteger& n2)
  {
    return cmp(N1, n2) <= 0;
  }
			

  // Comparison functions between a MachineInteger and a ZZ

  bool operator==(const MachineInteger& n1, const ZZ& N2)
  {
    return cmp(n1, N2) == 0;
  }


  bool operator!=(const MachineInteger& n1, const ZZ& N2)
  {
    return cmp(n1, N2) != 0;
  }


  bool operator> (const MachineInteger& n1, const ZZ& N2)
  {
    return cmp(n1, N2) > 0;
  }


  bool operator>=(const MachineInteger& n1, const ZZ& N2)
  {
    return cmp(n1, N2) >= 0;
  }


  bool operator< (const MachineInteger& n1, const ZZ& N2)
  {
    return cmp(n1, N2) < 0;
  }


  bool operator<=(const MachineInteger& n1, const ZZ& N2)
  {
    return cmp(n1, N2) <= 0;
  }
			


  //---------------------------------------------------------------------------
  // Assignment and assignment arithmetic functions

  ZZ& ZZ::operator+=(const ZZ& rhs)
  {
    mpz_add(myRep, myRep, mpzref(rhs));
    return *this;
  }


  ZZ& ZZ::operator-=(const ZZ& rhs)
  {
    mpz_sub(myRep, myRep, mpzref(rhs));
    return *this;
  }


  ZZ& ZZ::operator*=(const ZZ& rhs)
  {
    mpz_mul(myRep, myRep, mpzref(rhs));
    return *this;
  }


  ZZ& ZZ::operator/=(const ZZ& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::DivByZero, "ZZ /= ZZ");
    if (rhs < 0 && !mpz_divisible_p(myRep, mpzref(rhs)))
      CoCoA_ERROR(ERR::IntDivByNeg, "ZZ /= ZZ");
    mpz_fdiv_q(myRep, myRep, mpzref(rhs));
    return *this;
  }


  ZZ& ZZ::operator%=(const ZZ& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::ZeroModulus, "ZZ %= ZZ");
    if (rhs < 0 && !mpz_divisible_p(myRep, mpzref(rhs)))
      CoCoA_ERROR(ERR::IntDivByNeg, "ZZ %= ZZ");
    mpz_fdiv_r(myRep, myRep, mpzref(rhs));
    return *this;
  }


  //---------------------------------------------------------------------------
  // Assignment and assignment arithmetic with rhs a MachineInteger

  ZZ& ZZ::operator= (const MachineInteger& rhs)
  {
    if (IsNegative(rhs))
      mpz_set_si(myRep, AsSignedLong(rhs));
    else
      mpz_set_ui(myRep, AsUnsignedLong(rhs));

    return *this;
  }


  ZZ& ZZ::operator+=(const MachineInteger& rhs)
  {
    if (IsNegative(rhs))
      mpz_sub_ui(myRep, myRep, abs(rhs));
    else mpz_add_ui(myRep, myRep, AsUnsignedLong(rhs));
    return *this;
  }


  ZZ& ZZ::operator-=(const MachineInteger& rhs)
  {
    if (IsNegative(rhs))
      mpz_add_ui(myRep, myRep, abs(rhs));
    else mpz_sub_ui(myRep, myRep, AsUnsignedLong(rhs));
    return *this;
  }


  ZZ& ZZ::operator*=(const MachineInteger& rhs)
  {
    if (IsNegative(rhs))
      mpz_mul_si(myRep, myRep, AsSignedLong(rhs));
    else mpz_mul_ui(myRep, myRep, AsUnsignedLong(rhs));
    return *this;
  }


  ZZ& ZZ::operator/=(const MachineInteger& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::DivByZero, "ZZ /= MachineInteger");
    if (IsNegative(rhs))
    {
      if (!mpz_divisible_ui_p(myRep, abs(rhs)))
        CoCoA_ERROR(ERR::IntDivByNeg, "ZZ /= MachineInteger");
      mpz_neg(myRep, myRep);
      mpz_fdiv_q_ui(myRep, myRep, abs(rhs));
    }
    else
      mpz_fdiv_q_ui(myRep, myRep, AsUnsignedLong(rhs));
    return *this;
  }


  ZZ& ZZ::operator%=(const MachineInteger& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::ZeroModulus, "ZZ %= MachineInteger");
    if (IsNegative(rhs))
    {
      if (!mpz_divisible_ui_p(myRep, abs(rhs)))
        CoCoA_ERROR(ERR::IntDivByNeg, "ZZ %= MachineInteger");
      mpz_cdiv_r_ui(myRep, myRep, abs(rhs)); // least non-neg remainder!
    }
    else
      mpz_fdiv_r_ui(myRep, myRep, AsUnsignedLong(rhs));
    return *this;
  }



  //---------------------------------------------------------------------------
  // increment and decrement

  const ZZ& ZZ::operator++( )
  {
    mpz_add_ui(myRep, myRep, 1);
    return *this;
  }


  const ZZ& ZZ::operator--( )
  {
    mpz_sub_ui(myRep, myRep, 1);
    return *this;
  }


  const ZZ ZZ::operator++(int)  // INEFFICIENT
  {
    mpz_t ans;
    mpz_init_set(ans, myRep);
    ++*this;
    return ans[0];
  }


  const ZZ ZZ::operator--(int)  // INEFFICIENT
  {
    mpz_t ans;
    mpz_init_set(ans, myRep);
    --*this;
    return ans[0];
  }


  //---------------------------------------------------------------------------

  const ZZ::rtn abs(const ZZ& N)
  {
    mpz_t ans;
    mpz_init_set(ans, mpzref(N));
    mpz_abs(ans, ans);
    return ans[0];
  }


  void neg(ZZ& lhs, const ZZ& rhs)
  {
    mpz_neg(mpzref(lhs), mpzref(rhs));
  }


  void abs(ZZ& lhs, const ZZ& rhs)
  {
    mpz_abs(mpzref(lhs), mpzref(rhs));
  }


  //---------------------------------------------------------------------------
  // Efficient 3 argument arithmetic functions -- should be inline?

  void add(ZZ& lhs, const ZZ& N1, const ZZ& N2)
  {
    mpz_add(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void sub(ZZ& lhs, const ZZ& N1, const ZZ& N2)
  {
    mpz_sub(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void mul(ZZ& lhs, const ZZ& N1, const ZZ& N2)
  {
    mpz_mul(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void div(ZZ& lhs, const ZZ& N1, const ZZ& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "div(ZZ, ZZ, ZZ)");
    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "div(ZZ, ZZ, ZZ)");
    mpz_fdiv_q(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void mod(ZZ& lhs, const ZZ& N1, const ZZ& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::ZeroModulus, "mod(ZZ, ZZ, ZZ)");
    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "mod(ZZ, ZZ, ZZ)");
    mpz_fdiv_r(mpzref(lhs), mpzref(N1), mpzref(N2));  // least non-neg remainder
  }


  void quorem(ZZ& quo, ZZ& rem, const ZZ& num, const ZZ& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "quorem(ZZ, ZZ, ZZ, ZZ)");
    if (den < 0 && !mpz_divisible_p(mpzref(num), mpzref(den)))
      CoCoA_ERROR(ERR::IntDivByNeg, "quorem(ZZ, ZZ, ZZ, ZZ)");
    mpz_fdiv_qr(mpzref(quo), mpzref(rem), mpzref(num), mpzref(den));
  }


  void divexact(ZZ& lhs, const ZZ& N1, const ZZ& N2)
  {
    mpz_divexact(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  // Efficient 3 arg arithmetic where last arg is a MachineInteger

  void add(ZZ& lhs, const ZZ& N1, const MachineInteger& n2)
  {
    if (IsNegative(n2))
      mpz_sub_ui(mpzref(lhs), mpzref(N1), abs(n2));
    else
      mpz_add_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void sub(ZZ& lhs, const ZZ& N1, const MachineInteger& n2)
  {
    if (IsNegative(n2))
      mpz_add_ui(mpzref(lhs), mpzref(N1), abs(n2));
    else
      mpz_sub_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void mul(ZZ& lhs, const ZZ& N1, const MachineInteger& n2)
  {
    if (IsNegative(n2))
      mpz_mul_si(mpzref(lhs), mpzref(N1), AsSignedLong(n2));
    else
      mpz_mul_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void div(ZZ& lhs, const ZZ& N1, const MachineInteger& n2)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "div(ZZ, ZZ, MachineInteger)");
    if (IsNegative(n2))
    {
      if (!mpz_divisible_ui_p(mpzref(N1), abs(n2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "div(ZZ, ZZ, MachineInteger)");
      mpz_cdiv_q_ui(mpzref(lhs), mpzref(N1), abs(n2));
      mpz_neg(mpzref(lhs), mpzref(lhs));
    }
    else
      mpz_fdiv_q_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void mod(ZZ& lhs, const ZZ& N1, const MachineInteger& n2)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::ZeroModulus, "mod(ZZ, ZZ, MachineInteger)");
    if (IsNegative(n2))
    {
      if (!mpz_divisible_ui_p(mpzref(N1), abs(n2)))
        CoCoA_ERROR(ERR::IntDivByNeg, "mod(ZZ, ZZ, MachineInteger)");
      mpz_cdiv_r_ui(mpzref(lhs), mpzref(N1), abs(n2));
    }
    else
      mpz_fdiv_r_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void quorem(ZZ& quo, ZZ& rem, const ZZ& num, const MachineInteger& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "quorem(ZZ, ZZ, ZZ, MachineInteger)");
    if (IsNegative(den))
    {
      if (!mpz_divisible_ui_p(mpzref(num), abs(den)))
        CoCoA_ERROR(ERR::IntDivByNeg, "quorem(ZZ, ZZ, ZZ, MachineInteger)");
      mpz_cdiv_qr_ui(mpzref(quo), mpzref(rem), mpzref(num), abs(den));
      mpz_neg(mpzref(quo), mpzref(quo));
    }
    else
      mpz_fdiv_qr_ui(mpzref(quo), mpzref(rem), mpzref(num), AsUnsignedLong(den));
  }



  // Efficient 3 arg arithmetic where second arg is a MachineInteger

  void add(ZZ& lhs, const MachineInteger& n1, const ZZ& N2)
  {
    add(lhs, N2, n1);
  }


  void sub(ZZ& lhs, const MachineInteger& n1, const ZZ& N2)
  {
    sub(lhs, N2, n1);
    neg(lhs, lhs);
  }


  void mul(ZZ& lhs, const MachineInteger& n1, const ZZ& N2)
  {
    mul(lhs, N2, n1);
  }

// This function refers to a single "static" temporary variable of type
// ZZ, used only in div(ZZ,MachineInteger,ZZ), mod(ZZ,MachineInteger,ZZ),
// quorem(ZZ,ZZ,MachineInteger,ZZ).
// Using this ZZ_tmp function we only ever do at most one construction
// of an object of type ZZ, so the overhead should be pretty low.
  namespace
  {
    ZZ& ZZ_tmp()
    {
      static ZZ temporary;
      return temporary;
    }
  }

  void div(ZZ& lhs, const MachineInteger& n1, const ZZ& N2)
  {
    if (N2 == 0)
      CoCoA_ERROR(ERR::DivByZero, "div(ZZ, MachineInteger, ZZ)");
    ZZ_tmp() = n1;
    if (N2 < 0 && !mpz_divisible_p(mpzref(ZZ_tmp()), mpzref(N2)))
        CoCoA_ERROR(ERR::IntDivByNeg, "div(ZZ, MachineInteger, ZZ)");
    div(lhs, ZZ_tmp(), N2);
  }


  void mod(ZZ& lhs, const MachineInteger& n1, const ZZ& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::ZeroModulus, "mod(ZZ, MachineInteger, ZZ)");
    ZZ_tmp() = n1;
    if (N2 < 0 && !mpz_divisible_p(mpzref(ZZ_tmp()), mpzref(N2)))
        CoCoA_ERROR(ERR::IntDivByNeg, "mod(ZZ, MachineInteger, ZZ)");
    mod(lhs, ZZ_tmp(), N2);
  }


  void quorem(ZZ& quo, ZZ& rem, const MachineInteger& num, const ZZ& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "quorem(ZZ, ZZ, MachineInteger, ZZ)");
    ZZ_tmp() = num;
    if (den < 0 && !mpz_divisible_p(mpzref(ZZ_tmp()), mpzref(den)))
        CoCoA_ERROR(ERR::IntDivByNeg, "quorem(ZZ, ZZ, MachineInteger, ZZ)");
    quorem(quo, rem, ZZ_tmp(), den);
  }



  //---------------------------------------------------------------------------

//   const ZZ::rtn gcd(const ZZ& a, const ZZ& b)
//   {
//     //  if (IsZero(a) && IsZero(b)) throw invalid_argument("gcd(ZZ,ZZ): both arguments are 0");
//     mpz_t ans;
//     mpz_init(ans);
//     mpz_gcd(ans, mpzref(a), mpzref(b));
//     return ans[0];
//   }


  void gcd(ZZ& lhs, const ZZ& a, const ZZ& b)
  {
    //  if (IsZero(a) && IsZero(b)) throw invalid_argument("gcd(ZZ,ZZ,ZZ): both input arguments are 0");
    mpz_gcd(mpzref(lhs), mpzref(a), mpzref(b));
  }


//   const ZZ::rtn lcm(const ZZ& a, const ZZ& b)
//   {
//     mpz_t ans;
//     mpz_init(ans);
//     mpz_lcm(ans, mpzref(a), mpzref(b));
//     return ans[0];
//   }


  void lcm(ZZ& lhs, const ZZ& a, const ZZ& b)
  {
    mpz_lcm(mpzref(lhs), mpzref(a), mpzref(b));
  }


//   void gcdext(ZZ& g, ZZ& c_a, ZZ& c_b, const ZZ& a, const ZZ& b)
//   {
//     mpz_gcdext(mpzref(g), mpzref(c_a), mpzref(c_b), mpzref(a), mpzref(b));
//   }


//   bool invmod(ZZ& lhs, const ZZ& r, const ZZ& m)
//   {
//     return mpz_invert(mpzref(lhs), mpzref(r), mpzref(m));
//   }


  //---------------------------------------------------------------------------

//   void powermod(ZZ& lhs, const ZZ& base, const ZZ& exponent, const ZZ& modulus)
//   {
//     if (IsZero(modulus))
//       CoCoA_ERROR(ERR::ZeroModulus, "powermod(ZZ, ZZ, ZZ, ZZ)");
//     if (exponent < 0)
//       CoCoA_ERROR(ERR::NegExp, "powermod(ZZ, ZZ, ZZ, ZZ)");
//     mpz_powm(mpzref(lhs), mpzref(base), mpzref(exponent), mpzref(modulus));
//   }


//   void powermod(ZZ& lhs, const ZZ& base, const MachineInteger& exponent, const ZZ& modulus)
//   {
//     if (IsZero(modulus))
//       CoCoA_ERROR(ERR::ZeroModulus, "powermod(ZZ, ZZ, MachineInteger, ZZ)");
//     if (IsNegative(exponent))
//       CoCoA_ERROR(ERR::NegExp, "powermod(ZZ, ZZ, MachineInteger, ZZ)");
//     mpz_powm_ui(mpzref(lhs), mpzref(base), AsUnsignedLong(exponent), mpzref(modulus));
//   }


  const ZZ::rtn factorial(const ZZ& N)
  {
    if (N < 0)
      CoCoA_ERROR(ERR::BadArg, "factorial(ZZ)");
    unsigned long n;
    if (!convert(n, N))
      CoCoA_ERROR(ERR::ArgTooBig, "factorial(ZZ)");
    mpz_t ans;
    mpz_init(ans);
    mpz_fac_ui(ans, n);
    return ans[0];
  }


  const ZZ::rtn factorial(const MachineInteger& n)
  {
    if (IsNegative(n))
      CoCoA_ERROR(ERR::BadArg, "factorial(MachineInteger)");
    mpz_t ans;
    mpz_init(ans);
    mpz_fac_ui(ans, AsUnsignedLong(n));
    return ans[0];
  }


  const ZZ::rtn binomial(const ZZ& N, const ZZ& R)
  {
    mpz_t ans;
    mpz_init(ans);
    if (R < 0 || R > N)
      return ans[0]; // value is 0
    unsigned long r;
    if ((2*R < N && !convert(r, R)) || !convert(r, N-R))
      CoCoA_ERROR(ERR::ArgTooBig, "binomial(ZZ, ZZ)");
    mpz_bin_ui(ans, mpzref(N), r); // we know that r >= 0.
    return ans[0];
  }


  const ZZ::rtn binomial(const MachineInteger& n, const ZZ& R)
  {
    unsigned long r;
    if (R >= 0 && convert(r, R)) return binomial(n, r);
    // Either R < 0 or R > max unsigned long.
    if (R > 0 && IsNegative(n))
      CoCoA_ERROR(ERR::ArgTooBig, "binomial(MachineInteger, ZZ)");
    // R is negative, or n is positive (and necessarily smaller than R), so ans is 0.
    mpz_t ans;
    mpz_init(ans);
    return ans[0]; // value is 0
  }


  const ZZ::rtn binomial(const ZZ& N, const MachineInteger& r)
  {
    mpz_t ans;
    mpz_init(ans);
    if (!IsNegative(r))
      mpz_bin_ui(ans, mpzref(N), AsUnsignedLong(r)); // compute ans only if r >= 0, o/w ans stays 0.
    return ans[0];
  }


  const ZZ::rtn binomial(const MachineInteger& n, const MachineInteger& r)
  {
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(r)) return ans[0]; // value is 0
    if (IsNegative(n))
    {
      ZZ N(abs(n));
      N += AsUnsignedLong(r)-1;
      mpz_bin_ui(ans, mpzref(N), AsUnsignedLong(r));
      if (IsOdd(r)) mpz_neg(ans, ans);
    }
    else mpz_bin_uiui(ans, AsUnsignedLong(n), AsUnsignedLong(r));
    return ans[0];
  }


  const ZZ::rtn fibonacci(const ZZ& N)
  {
    unsigned long n;
    if (!convert(n, N))
      CoCoA_ERROR(ERR::ArgTooBig, "fibonacci(ZZ)");
    return fibonacci(n);
  }


  const ZZ::rtn fibonacci(const MachineInteger& n)
  {
    bool EvenNegative = IsNegative(n) && IsEven(n);
    mpz_t ans;
    mpz_init(ans);
    unsigned long arg = abs(n);
    mpz_fib_ui(ans, arg);
    if (EvenNegative) mpz_neg(ans, ans);
    return ans[0];
  }

  //---------------------------------------------------------------------------

  const ZZ::rtn operator+(const ZZ& N1, const MachineInteger& n2)
  {
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n2))
      mpz_sub_ui(ans, mpzref(N1), abs(n2));
    else
      mpz_add_ui(ans, mpzref(N1), AsUnsignedLong(n2));
    return ans[0];
  }


  const ZZ::rtn operator-(const ZZ& N1, const MachineInteger& n2)
  {
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n2))
      mpz_add_ui(ans, mpzref(N1), abs(n2));
    else
      mpz_sub_ui(ans, mpzref(N1), AsUnsignedLong(n2));
    return ans[0];
  }


  const ZZ::rtn operator*(const ZZ& N1, const MachineInteger& n2)
  {
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n2))
      mpz_mul_si(ans, mpzref(N1), AsSignedLong(n2));
    else
      mpz_mul_ui(ans, mpzref(N1), AsUnsignedLong(n2));
    return ans[0];
  }


  const ZZ::rtn operator/(const ZZ& N1, const MachineInteger& n2)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "ZZ / MachineInteger");
    if (IsNegative(n2) && !mpz_divisible_ui_p(mpzref(N1), abs(n2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "ZZ / MachineInteger");
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n2))
    {
      mpz_cdiv_q_ui(ans, mpzref(N1), abs(n2));
      mpz_neg(ans, ans);
    }
    else
      mpz_fdiv_q_ui(ans, mpzref(N1), AsUnsignedLong(n2));
    return ans[0];
  }


  unsigned long operator%(const ZZ& N1, const MachineInteger& n2)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "ZZ % MachineInteger");
    if (IsNegative(n2) && !mpz_divisible_ui_p(mpzref(N1), abs(n2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "ZZ % MachineInteger");
    return mpz_fdiv_ui(mpzref(N1), AsUnsignedLong(n2));
//     mpz_t ans;
//     mpz_init(ans);
//     if (IsNegative(n2))
//     {
//       mpz_cdiv_r_ui(ans, mpzref(N1), abs(n2));
//     }
//     else
//       mpz_fdiv_r_ui(ans, mpzref(N1), AsUnsignedLong(n2));
//     return ans[0];
  }


  const ZZ::rtn operator+(const MachineInteger& n1, const ZZ& N2)
  {
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n1))
      mpz_sub_ui(ans, mpzref(N2), abs(n1));
    else
      mpz_add_ui(ans, mpzref(N2), AsUnsignedLong(n1));
    return ans[0];
  }


  const ZZ::rtn operator-(const MachineInteger& n1, const ZZ& N2)
  {
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n1))
      mpz_add_ui(ans, mpzref(N2), abs(n1));
    mpz_sub_ui(ans, mpzref(N2), AsUnsignedLong(n1));
    mpz_neg(ans, ans);
    return ans[0];
  }


  const ZZ::rtn operator*(const MachineInteger& n1, const ZZ& N2)
  {
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n1))
      mpz_mul_si(ans, mpzref(N2), AsSignedLong(n1));
    else
      mpz_mul_ui(ans, mpzref(N2), AsUnsignedLong(n1));
    return ans[0];
  }


  const ZZ::rtn operator/(const MachineInteger& n1, const ZZ& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "MachineInteger / ZZ");
    ZZ_tmp() = n1;
    if (N2 < 0 && !mpz_divisible_p(mpzref(ZZ_tmp()), mpzref(N2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "MachineInteger / ZZ");
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n1))
      mpz_set_si(ans, AsSignedLong(n1));
    else
      mpz_set_ui(ans, AsUnsignedLong(n1));
    mpz_fdiv_q(ans, ans, mpzref(N2));
    return ans[0];
  }


  const ZZ::rtn operator%(const MachineInteger& n1, const ZZ& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "MachineInteger % ZZ");
    ZZ_tmp() = n1;
    if (N2 < 0 && !mpz_divisible_p(mpzref(ZZ_tmp()), mpzref(N2)))
      CoCoA_ERROR(ERR::IntDivByNeg, "MachineInteger % ZZ");
    // BUG/SLUG this could be made rather more efficient
    mpz_t ans;
    mpz_init(ans);
    if (IsNegative(n1))
      mpz_set_si(ans, AsSignedLong(n1));
    else
      mpz_set_ui(ans, AsUnsignedLong(n1));
    mpz_fdiv_r(ans, ans, mpzref(N2));
    return ans[0];
  }


} // end of namespace CoCoA



// The next few lines contain RCS header/log information.
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ZZ.C,v 1.12 2009/06/05 12:08:27 abbott Exp $
// $Log: ZZ.C,v $
// Revision 1.12  2009/06/05 12:08:27  abbott
// Changed return type of operator%(ZZ,MachineInteger); it is now unsigned long
// instead of ZZ.
//
// Revision 1.11  2009/01/12 14:34:10  abbott
// Corrected IsDigitBase (following a bug report from anonymous)
// for octal and hexadecimal digits.  Added 3 lines of comment too.
//
// Revision 1.10  2008/12/16 21:14:47  abbott
// In functions taking a machine integer changed arg type from MachineInteger
// to const-ref-MachineInteger.
//
// Revision 1.9  2008/11/18 10:25:48  abbott
// Added function round.
//
// Revision 1.8  2008/04/22 13:09:16  abbott
// Removed IsPositive and IsNegative functions for ZZs.
// Comparison between RingElem and 0 is now handled specially (specially fast?).
//
// Revision 1.7  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.6  2007/06/01 13:56:08  abbott
// Now gives error whenever you try to compute 0^0 -- previously some cases were not checked.
//
// Revision 1.5  2007/05/22 22:51:39  abbott
// Changed name of fn ndigits to NumDigits.
// Changed return type of exponent and NumDigits.
// Changed some exceptions from nonstandard to the appropriate standard one.
//
// Revision 1.4  2007/05/21 12:57:28  abbott
// New class for passing machine integers as args; includes some simple
// operations on machine integers (cmp, gcd, IsNegative,...).  Operations
// between ZZ and machine integers modified to use the new class.  Inexact
// integer division (of a ZZ) by a negative value now triggers an error;
// new error for reporting inexact integer division by a negative value.
//
// Revision 1.3  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.2  2007/03/16 17:43:05  abbott
// Added new convert function (from ZZ to double).
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/08 18:22:28  cocoa
// Just whitespace cleaning.
//
// Revision 1.8  2007/01/13 14:14:34  cocoa
// Overhaul of RingHom code: it nows uses SmartPtrIRC, and printing is more logical.
// Have not yet updated the documentation.
//
// Revision 1.7  2007/01/09 15:52:08  cocoa
// Changed QBGenerator to use std::vector instead of std::list for the result.
// Minor mod to configure script.
//
// Revision 1.6  2006/11/29 11:59:35  cocoa
// -- fixed: convert(double& z, const ZZ& num, const ZZ& den) now returns
//    bool (was void) and does not throw errors
//
// Revision 1.5  2006/11/27 13:06:22  cocoa
// Anna and Michael made me check without writing a proper message.
//
// Revision 1.4  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.3  2006/10/06 09:36:13  cocoa
// Fixed two minor bugs introduced at last check in.
//
// Revision 1.2  2006/09/29 11:46:54  cocoa
// Corrected bug in convert(ZZ, ZZ, double) -- now correct and simpler.
// Previously went into infinite loop on negative doubles.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.5  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.4  2006/03/27 12:21:25  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.3  2006/03/14 15:01:49  cocoa
// Improved the implementation of ring member fns for computing powers.
// Should keep Intel C++ compiler quieter too.
//
// Revision 1.2  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.2  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.3  2005/04/20 15:40:47  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.2  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.8  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.7  2004/11/05 15:36:57  cocoa
// Removed call to new(nothrow) inside a convert function.  Replaced
// it by use of a std::vector -- code is certainly cleaner and clearer
// now, but possibly with an inconsequential reduction in run-time
// performance.
//
// Revision 1.6  2004/11/04 18:47:42  cocoa
// (1) Ring member functions which previously expected mpz_t args
//     now expect ZZ args.  Numerous minor consequential changes.
// (2) Renamed function which gives access to the mpz_t value inside
//     a ZZ object: previously was raw(...), now is mpzref(...).
//     Plenty of calls had to be altered.
//
// Revision 1.5  2004/05/27 16:14:02  cocoa
// Minor revision for new coding conventions.
//
// Revision 1.4  2004/05/24 15:52:13  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.3  2004/02/03 16:16:20  cocoa
// Removed pointless IamGCDDomain functions from several concrete rings.
// Added IamOrderedDomain functions where appropriate.
// Tidied ctors for the small finite fields.
//
// Revision 1.2  2003/10/17 10:51:06  cocoa
// Major cleaning, and new naming convention.
//
// Revision 1.1.1.1  2003/09/24 12:55:43  cocoa
// Imported files
//
// Revision 1.2  2003/06/23 16:13:01  abbott
// Minor cleaning prior to public release.
//
// Revision 1.1  2003/05/14 17:12:40  abbott
// Initial revision
//
