//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/convert.H"
#include "CoCoA/ZZ.H"

//#include <string>
using std::string;
#include <vector>
using std::vector;     // Used only in converting to a string.
#include <new>
using std::bad_alloc;  // Used only in converting to a string.
#include <limits>
using std::numeric_limits;
#include <cmath>
using std::ldexp;

namespace CoCoA
{

  bool convert(long& lhs, const ZZ& src)
  {
    if (!mpz_fits_slong_p(mpzref(src))) return false;
    lhs = mpz_get_si(mpzref(src));
    return true;
  }


  bool convert(int& lhs, const ZZ& src)
  {
    if (!mpz_fits_sint_p(mpzref(src))) return false;
    lhs = mpz_get_si(mpzref(src));
    return true;
  }


  bool convert(unsigned long& lhs, const ZZ& src)
  {
    if (!mpz_fits_ulong_p(mpzref(src))) return false;
    lhs = mpz_get_ui(mpzref(src));
    return true;
  }


  bool convert(unsigned int& lhs, const ZZ& src)
  {
    if (!mpz_fits_uint_p(mpzref(src))) return false;
    lhs = mpz_get_ui(mpzref(src));
    return true;
  }


  bool convert(string& lhs, const ZZ& src, int base)
  {
    if (base < 2 || base > 36)
      CoCoA_ERROR(ERR::BadNumBase, "convert(string,ZZ,int)");
    const size_t digits = NumDigits(src, base);
    try
    {
      lhs.reserve(digits+1); // +1 to allow to minus sign
      vector<char> buffer(digits+2); // +2 to allow for minus sign and terminating NUL
      mpz_get_str(&buffer[0], base, mpzref(src));
      lhs = &buffer[0]; // Won't throw as space is already reserved.
      return true;
    }
    catch (const std::bad_alloc&)  // In case we run out of memory.
    { return false; }
  }


  bool convert(ZZ& lhs, const string& src, int base)
  {
    if (base != 0 && (base < 2 || base > 36))
      CoCoA_ERROR(ERR::BadNumBase, "convert(ZZ,string,int)");
    return !mpz_set_str(mpzref(lhs), src.c_str(), base);
  }



  // Sets z = N if possible.
  bool convert(double& z, const ZZ& N)
  {
    if (IsZero(N)) { z = 0.0; return true; }
    // I prefer to do this manually because mpz_get_d does not give
    // good guarantees about what happens when ovverflow occurs.
    long ExpN;
    double MantissaN = mpz_get_d_2exp(&ExpN, mpzref(N));
    if (ExpN > numeric_limits<double>::max_exponent)
    {
      return false;
    }
    z = ldexp(MantissaN, ExpN);
    return true;
  }



  bool convert(ZZ& num, ZZ& den, double z)  // NB always returns true, NOT exception clean
  {
    //??? BUG should handle infinity and Nan???
    if (z == 0) { num = 0; den = 1; return true; }
    mpq_t rational;
    mpq_init(rational);
    mpq_set_d(rational, z);  // Don't know what happens if this fails.
    mpz_swap(mpzref(num), mpq_numref(rational));
    mpz_swap(mpzref(den), mpq_denref(rational));
    mpq_clear(rational);
    return true;
  }


  // Sets z = num/den if possible.
  // Underflow sets z to 0 and returns true.
  // Overflow returns false (and does not change z).
  bool convert(double& z, const ZZ& num, const ZZ& den)
  {
    if (IsZero(den)) { return false; }
    if (IsZero(num)) { z = 0; return true; }
    // I prefer to do this manually rather than use mpq_get_d because
    // num and den may not be coprime, and the cost of computing the
    // gcd could be high.  Also mpq_get_d does not give nice guarantees
    // about what happens in the case of over-/under-flow.
    long Nexp;
    double N = mpz_get_d_2exp(&Nexp, mpzref(num));
    long Dexp;
    double D = mpz_get_d_2exp(&Dexp, mpzref(den));
    long ExpAns = Nexp-Dexp;
    double MantissaAns = N/D;
    if (MantissaAns >= 1 || MantissaAns <= -1)
    { MantissaAns /= 2; ++ExpAns; }

    // I believe this is the right interpretation of the exponent limits...
    if (ExpAns < numeric_limits<double>::min_exponent)
    {
      // We have underflow, so convert to 0 and indicate "success"
      z = 0.0;
      return true;
    }

    if (ExpAns > numeric_limits<double>::max_exponent)
    {
      // We have overflow, so indicate "failure".
      return false;
    }

    z = ldexp(MantissaAns, ExpAns); // Cannot overflow or underflow.
    return true;
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/convert.C,v 1.5 2008/10/07 15:45:22 abbott Exp $
// $Log: convert.C,v $
// Revision 1.5  2008/10/07 15:45:22  abbott
// Changed ErrorInfo objects so they include the name of their own error ID.
// Changed catch statements to catch const objects.
// Removed calls to the member fn which accessed the error ID member of an
// ErrorInfo; now you simply compare directly with the error ID (makes the
// code easier to read).
//
// Revision 1.4  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.3  2007/05/22 22:51:39  abbott
// Changed name of fn ndigits to NumDigits.
// Changed return type of exponent and NumDigits.
// Changed some exceptions from nonstandard to the appropriate standard one.
//
// Revision 1.2  2007/03/28 15:34:03  abbott
// Underflow during conversion to a double no longer causes "failure" but now
// converts the value to zero and yields "success".
//
// Revision 1.1  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.3  2007/03/03 15:24:39  cocoa
// Changed 2006 to 2007.
//
// Revision 1.2  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
