//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/GlobalManager.H"
#include "CoCoA/BuildInfo.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/FreeModule.H"
#include "CoCoA/TmpGOperations.H"
#include "CoCoA/TmpGPoly.H"
#include "CoCoA/ModuleTermOrdering.H"
#include "CoCoA/RingDistrMPoly.H"
#include "CoCoA/RingDistrMPolyInlFpPP.H"
#include "CoCoA/RingDistrMPolyInlPP.H"
#include "CoCoA/RingFp.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/io.H"
#include "CoCoA/submodule.H"
#include "CoCoA/symbol.H"

using namespace CoCoA;

#include <iostream>
using std::endl;
#include <vector>
using std::vector;
#include <string>

//----------------------------------------------------------------------
// Test for Buchberger's Algorithm 
// functions: ComputeGBasis
// environments: DMP (Fp, Q), DMPI (Fp), DMPII
//----------------------------------------------------------------------

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)

// expecting P to have 5 indeterminates
void testC4_h(SparsePolyRing P, const std::string& S)
{
  GlobalOutput() << "Executing test C4_h "+S<<endl;
  RingElem x = indet(P,0),  y = indet(P,1),  z = indet(P,2), t=indet(P,3);
  RingElem h = indet(P,4);
  PolyList InputPolys,GB;
  InputPolys.push_back(x+y+z+t);
  InputPolys.push_back(x*y+y*z+z*t+t*x);
  InputPolys.push_back(x*y*z+y*z*t+z*t*x+t*x*y);
  InputPolys.push_back(x*y*z*t-power(h,4));
  ComputeGBasis(GB, InputPolys);
  monic(GB);
  GlobalOutput() << "C4_h = " << GB << "\n"<<endl;
}


void testC4(SparsePolyRing P, const std::string& S)
{
  GlobalOutput() << "Executing test C4 ";
  if (GradingDim(P)!=0)
    GlobalOutput()<<"Graded Ring ";
  else
    GlobalOutput()<<"Affine Ring ";
  GlobalOutput()<<"Inhomogenous Input "+S<<endl;
  RingElem x = indet(P,0),  y = indet(P,1),  z = indet(P,2), t=indet(P,3);
  PolyList InputPolys,GB;
  InputPolys.push_back(x+y+z+t);
  InputPolys.push_back(x*y+y*z+z*t+t*x);
  InputPolys.push_back(x*y*z+y*z*t+z*t*x+t*x*y);
  InputPolys.push_back(x*y*z*t-1);
  ComputeGBasis(GB, InputPolys);
  monic(GB);
  GlobalOutput() << "C4 = " << GB << "\n"<<endl;
}

// expecting P to have 4 indeterminates
void testCurve_h(SparsePolyRing P, const std::string& S)
{
  GlobalOutput() << "Executing test Curve_h - AllGraded "+S<<endl;
  RingElem t = indet(P,0),  x = indet(P,1),  y = indet(P,2), z=indet(P,3);
  PolyList InputPolys,GB;
  InputPolys.push_back(power(t,3)-power(x,3));
  InputPolys.push_back(power(t,5)-power(y,5));
  InputPolys.push_back(power(t,7)-power(z,7));
  ComputeGBasis(GB, InputPolys);
  monic(GB);
  GlobalOutput() << "Curve_h = " << GB << "\n"<<endl;
}

// expecting P to have 5 indeterminates
void testCurve(SparsePolyRing P, const std::string& S)
{
  GlobalOutput() << "Executing test Curve "<<endl;
  if (GradingDim(P)!=0)
    GlobalOutput()<<"Graded Ring ";
  else
    GlobalOutput()<<"Affine Ring ";
  GlobalOutput()<<"Inhomogenous Input "+S<<endl;
  RingElem t = indet(P,0),  x = indet(P,1),  y = indet(P,2), z=indet(P,3);
  PolyList InputPolys,GB;
  InputPolys.push_back(power(t,3)-x);
  InputPolys.push_back(power(t,5)-y);
  InputPolys.push_back(power(t,7)-z);
  ComputeGBasis(GB, InputPolys);
  monic(GB);
  GlobalOutput() << "Curve = " << GB << "\n"<<endl;
}

void program()
{
  GlobalManager CoCoAFoundations(UseNonNegResidues);

  // This is a test of the correctness of GB computations for the three cases:
  // AllGraded, AllAffine and GradedRing, Inhomogenous Input for ideals only.
  // This does not check if the computations have been done in the right way,
  // only if the result is correct.

  // the results have been tested with cocoa 4
  
  ring ZMod = NewZmod(101);  
  
  // Cyclic 4
  
  const vector<symbol> X = SymbolRange("x", 0, 4);
  SparsePolyRing P(NewPolyRing_DMPI(ZMod, X));
  SparsePolyRing Q(NewPolyRing_DMPII(ZMod, X));
  SparsePolyRing R(NewPolyRing(RingQ(), 5));
  SparsePolyRing T(NewPolyRing_DMPI(RingQ(), X));

  // Graded Ring, Homogeneous Input
  testC4_h(P,"NewPolyRing_DMPI Fp");
  testC4_h(Q,"NewPolyRing_DMPII Fp");
  testC4_h(R,"NewPolyRing Q");
  testC4_h(T,"NewPolyRing_DMPI Q");
 
  // Graded Ring, Inhomogeneous Input
  testC4(P,"NewPolyRing_DMPI Fp");
  testC4(Q,"NewPolyRing_DMPII Fp");
  testC4(R,"NewPolyRing Q");
  testC4(T,"NewPolyRing_DMPI Q");
  
  SparsePolyRing LexR = NewPolyRing(RingQ(),X,NewLexOrdering(5));

  // Affine Ring, Inhomogeneous Input
  //testC4(LexR,"NewPolyRing Q"); 
    
    
  // Curve

  const vector<symbol> Y = SymbolRange("x", 0 ,3);
  SparsePolyRing P1(NewPolyRing_DMPI(ZMod, Y));
  SparsePolyRing Q1(NewPolyRing_DMPII(ZMod, Y));
  SparsePolyRing R1(NewPolyRing(RingQ(), 4));
  SparsePolyRing T1 (NewPolyRing_DMPI(RingQ(), Y));
 
    // Graded Ring, Homogeneous Input
  testCurve_h(P1,"NewPolyRing_DMPI Fp");
  testCurve_h(Q1,"NewPolyRing_DMPII Fp");
  testCurve_h(R1,"NewPolyRing Q");
  testCurve_h(T1,"NewPolyRing_DMPI Q");
  
    //Graded Ring, Inhomogeneous Input
  testCurve(P1,"NewPolyRing_DMPI Fp");
  testCurve(Q1,"NewPolyRing_DMPII Fp");
  testCurve(R1,"NewPolyRing Q");
  testCurve(T1,"NewPolyRing_DMPI Q");
  
  SparsePolyRing LexR1 = NewPolyRing(RingQ(),Y,NewLexOrdering(4));

  // Affine Ring, Inhomogeneous Input
  testCurve(LexR1,"NewPolyRing Q");
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
